# -*- coding: utf-8 -*-
"""
transaction contexts module.
"""

import pyrin.database.services as database_services

from pyrin.core.contexts import ContextManagerBase


class TransactionalContextManagerBase(ContextManagerBase):
    """
    transactional context manager base class.

    this class should be used as the base for all transactional context managers.
    """

    def __init__(self, store, **options):
        """
        initializes an instance of TransactionalContextManagerBase.

        :param CoreSession store: the session object for current context.
        """

        self._store = store

    def __enter__(self):
        """
        begins the current context and returns the related session for current context.

        note that you *should not* commit, flush or rollback the transaction inside
        the current context. the context manager itself will do it automatically.
        if you do commit, flush or rollback manually, unexpected behaviors may occur.

        :rtype: CoreSession
        """

        return self._store

    def __exit__(self, exc_type, exc_value, traceback):
        """
        does the finalizing of current transaction.

        it commits the transaction if it was successful, otherwise rollbacks it.

        :param type[Exception] exc_type: the exception type that has been
                                         occurred during current context.

        :param Exception exc_value: exception instance that has been
                                    occurred during current context.

        :param traceback traceback: traceback of occurred exception.
        """

        try:
            if exc_type is None:
                try:
                    self._store.commit()
                except Exception as error:
                    self._store.rollback()
                    raise error
            else:
                self._store.rollback()
        except Exception as error:
            if self._should_be_raised(error, exc_type) is True:
                raise error


class atomic_context(TransactionalContextManagerBase):
    """
    atomic context manager to make a code block execution atomic.

    meaning that before starting the execution of the code, a new session with a
    new transaction will be started, and after the completion of that code, if it
    was successful, the new transaction will be committed or if it was not successful
    the new transaction will be rolled back without the consideration or affecting the
    parent transaction which by default is scoped to request. the corresponding new
    session will also be removed after code execution.

    note that you *should not* commit, flush or rollback the transaction inside
    the current context. the context manager itself will do it automatically.
    if you do commit, flush or rollback manually, unexpected behaviors may occur.

    also note that you *should not* remove the corresponding session from session factory
    when using `atomic_context`. the removal operation will be handled by context manager
    itself and if you remove session manually, it will cause broken chain of sessions
    and unexpected behaviour.

    this context manager also supports multiple `atomic_context` usage in a single
    call hierarchy.

    for example:

    def service_root():
        store = get_current_store()
        value = EntityRoot()
        store.add(value)
        service_a()

    def service_a():
        with atomic_context() as store:
            value = EntityA()
            store.add(value)
            service_b()
            raise Exception()

    def service_b():
        with atomic_context() as store:
            value = EntityB()
            store.add(value)
            service_c()

    def service_c():
        with atomic_context():
            value = EntityC()
            value.save()

    in the above example, if the call hierarchy starts with `service_root()`, at
    the end, the data of `service_c` and `service_b` will be persisted into database.
    but the data of `service_a` and `service_root` will not be persisted because
    `service_a` raises an error before finish.
    """

    def __init__(self, **options):
        """
        initializes an instance of atomic_context.

        :keyword bool expire_on_commit: expire atomic session after commit.
                                        it is useful to set it to True if
                                        the atomic function does not return
                                        any entities for post-processing.
                                        defaults to False if not provided.
        """

        super().__init__(database_services.get_atomic_store(**options))

    def __exit__(self, exc_type, exc_value, traceback):
        """
        does the finalizing of atomic transaction.

        it commits the transaction if it was successful, otherwise rollbacks it.

        :param type[Exception] exc_type: the exception type that has been
                                         occurred during current context.

        :param Exception exc_value: exception instance that has been
                                    occurred during current context.

        :param traceback traceback: traceback of occurred exception.
        """

        try:
            try:
                super().__exit__(exc_type, exc_value, traceback)
            finally:
                self._remove_session()
        except Exception as error:
            if self._should_be_raised(error, exc_type) is True:
                raise error

    def _remove_session(self):
        """
        removes the current atomic session from session factory.
        """

        factory = database_services.get_current_session_factory()
        factory.remove(atomic=True)


class nested_context(TransactionalContextManagerBase):
    """
    nested context manager to make a code block executed in nested transaction.

    meaning that before starting the execution of the code, a new nested transaction
    will be started, and after the completion of that code, if it was not successful
    the nested transaction will be rolled back without the consideration or affecting
    the parent transaction which by default is scoped to request.

    note that in nested transactions, the parent transaction must be committed to persist
    generated data during nested transaction, so committing the nested transaction itself,
    does not persist anything, it just releases the savepoint. if you want an independent
    transaction from parent that could commit its own changes on its own, use `atomic_context`.

    note that you *should not* commit, flush or rollback the transaction inside
    the current context. the context manager itself will do it automatically.
    if you do commit, flush or rollback manually, unexpected behaviors may occur.

    this context manager also supports multiple `nested_context` usage in a single
    call hierarchy.

    example usage:

    with nested_context():
        store = database_services.get_current_store()
        user = UserEntity(id=1, name='Dave')
        store.add(user)
    """

    def __init__(self):
        """
        initializes an instance of nested_context.
        """

        super().__init__(database_services.get_current_store().begin_nested())

    def __enter__(self):
        """
        begins the current context.

        note that you *should not* commit, flush or rollback the transaction inside
        the current context. the context manager itself will do it automatically.
        if you do commit, flush or rollback manually, unexpected behaviors may occur.

        also note that by beginning the context, it does not return any session
        object. so you should not use `with nested_context() as store` code style.
        because store will always be None.

        :rtype: CoreSession
        """

        return None


class subtransaction_context(TransactionalContextManagerBase):
    """
    subtransaction context manager to make a code block executed in subtransaction.

    meaning that before starting the execution of the code, a new subtransaction
    will be started, and after the completion of that code, if it was not successful
    the subtransaction will be rolled back and it will also force the parent transaction
    which by default is scoped to request to be rolled backed.

    note that in subtransactions, the parent transaction must be committed to persist
    generated data during subtransaction, so committing the subtransaction itself, does
    not persist anything, it just ends the subtransaction's scope. if you want an independent
    transaction from parent that could commit its own changes on its own, use `atomic_context`.

    note that you *should not* commit, flush or rollback the transaction inside
    the current context. the context manager itself will do it automatically.
    if you do commit, flush or rollback manually, unexpected behaviors may occur.

    this context manager also supports multiple `subtransaction_context` usage in a single
    call hierarchy.

    example usage:

    with subtransaction_context():
        store = database_services.get_current_store()
        user = UserEntity(id=1, name='Dave')
        store.add(user)
    """

    def __init__(self):
        """
        initializes an instance of subtransaction_context.
        """

        super().__init__(database_services.get_current_store().begin(subtransactions=True))

    def __enter__(self):
        """
        begins the current context.

        note that you *should not* commit, flush or rollback the transaction inside
        the current context. the context manager itself will do it automatically.
        if you do commit, flush or rollback manually, unexpected behaviors may occur.

        also note that by beginning the context, it does not return any session
        object. so you should not use `with subtransaction_context() as store` code
        style. because store will always be None.

        :rtype: CoreSession
        """

        return None


class transient_context(atomic_context):
    """
    transient context manager to make a code block execution transient.

    meaning that before starting the execution of the code block, a new session with a
    new transaction will be started, and after the completion of that code, the
    new transaction will be rolled back without the consideration or affecting the
    parent transaction which by default is scoped to request. the corresponding new
    session will also be removed after code execution.

    note that you *should not* commit, flush or rollback anything inside a transient
    block, the transient context manager will handle rollback operation when needed.
    otherwise, unexpected behaviors may occur.

    also note that you *should not* remove the corresponding session from session factory
    when using transient context manager. the removal operation will be handled by decorator
    itself and if you remove session manually, it will cause broken chain of sessions
    and unexpected behaviour.

    this context manager also supports multiple `transient_context` usage in a single
    call hierarchy.

    for example:

    def service_root():
        store = get_current_store()
        value = EntityRoot()
        store.add(value)
        service_a()

    def service_a():
        with atomic_context() as store:
            value = EntityA()
            store.add(value)
            service_b()

    def service_b():
        with transient_context() as store:
            value = EntityB()
            store.add(value)
            service_c()

    def service_c():
        with transient_context():
            value = EntityC()
            value.save()

    in the above example, if the call hierarchy starts with `service_root()`, at
    the end, the data of `service_root` and `service_a` will be persisted into database.
    but the data of `service_b` and `service_c` will not be persisted because
    they are marked as transient.
    """

    def __exit__(self, exc_type, exc_value, traceback):
        """
        does the finalizing of transient transaction.

        it rollbacks the transaction at the end.

        :param type[Exception] exc_type: the exception type that has been
                                         occurred during current context.

        :param Exception exc_value: exception instance that has been
                                    occurred during current context.

        :param traceback traceback: traceback of occurred exception.
        """

        try:
            try:
                self._store.rollback()
            finally:
                self._remove_session()
        except Exception as error:
            if self._should_be_raised(error, exc_type) is True:
                raise error



# -*- coding: utf-8 -*-
"""
cli core manager module.
"""

from pyrin.cli.core import CLICorePackage
from pyrin.cli.core.exceptions import CLICoreTemplateHandlerNotFoundError
from pyrin.cli.core.template.handlers import ScriptsTemplateHandler, ApplicationTemplateHandler
from pyrin.core.structs import Manager, DTO


class CLICoreManager(Manager):
    """
    cli core manager class.
    """

    package_class = CLICorePackage

    def __init__(self):
        """
        initializes an instance of CLICoreManager.
        """

        super().__init__()

        scripts_handler = ScriptsTemplateHandler()
        application_handler = ApplicationTemplateHandler()
        scripts_handler.set_next(application_handler)

        self._handlers = DTO(project=[scripts_handler, application_handler])

    def create(self, handler_name):
        """
        creates the required templates using relevant handlers.

        :param str handler_name: handler name to be used.

        :raises CLICoreTemplateHandlerNotFoundError: cli core template handler not found error.
        """

        handler = self._get_handler(handler_name)
        handler.create()

    def _get_handler(self, name):
        """
        gets the requested handler with given name.

        :param str name: handler name to be get.

        :raises CLICoreTemplateHandlerNotFoundError: cli core template handler not found error.

        :rtype: ProjectStructureTemplateHandlerBase
        """

        if name not in self._handlers:
            raise CLICoreTemplateHandlerNotFoundError('Command [{name}] is not valid. '
                                                      'available commands: {commands}.'
                                                      .format(name=name,
                                                              commands=
                                                              list(self._handlers.keys())))

        return self._handlers[name][0]



# -*- coding: utf-8 -*-
"""
encryption component module.
"""

from pyrin.application.decorators import component
from pyrin.security.encryption import EncryptionPackage
from pyrin.security.encryption.manager import EncryptionManager
from pyrin.application.structs import Component


@component(EncryptionPackage.COMPONENT_NAME)
class EncryptionComponent(Component, EncryptionManager):
    """
    encryption component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
caching local items permanent module.
"""

from pyrin.caching.local.items.base import LocalCacheItemBase


class PermanentLocalCacheItem(LocalCacheItemBase):
    """
    permanent local cache item class.

    this type of cache item does not support expire time.
    it also keeps the original value into the cache to gain performance.
    """
    pass



# -*- coding: utf-8 -*-
"""
converters package.
"""

from pyrin.packaging.base import Package


class ConvertersPackage(Package):
    """
    converters package class.
    """

    NAME = __name__
    DEPENDS = []



# -*- coding: utf-8 -*-
"""
serializer manager module.
"""

from pyrin.converters.serializer import SerializerPackage
from pyrin.converters.serializer.interface import AbstractSerializerBase
from pyrin.core.structs import Context, Manager
from pyrin.core.globals import NULL
from pyrin.utils.custom_print import print_warning
from pyrin.converters.serializer.exceptions import InvalidSerializerTypeError, \
    DuplicatedSerializerError


class SerializerManager(Manager):
    """
    serializer manager class.
    """

    package_class = SerializerPackage

    def __init__(self):
        """
        initializes an instance of SerializerManager.
        """

        super().__init__()

        # a dictionary containing information of registered serializers.
        # example: dic(type accepted_type: AbstractSerializerBase instance)
        self._serializers = Context()

    def serialize(self, value, **options):
        """
        serializes the given value.

        returns serialized value on success or the
        same input value if serialization fails.

        all extra keyword arguments will be passed to underlying serializer.

        :param object value: value to be serialized.

        :keyword ResultSchema result_schema: result schema instance to be
                                             used to create computed columns.
                                             defaults to None if not provided.

        :returns: serialized object
        :rtype: dict | list[dict]
        """

        if value is None:
            return value

        serializers = self.get_serializers(type(value))
        for item in serializers:
            serialized_value = item.serialize(value, **options)
            if serialized_value is not NULL:
                return serialized_value

        return value

    def register_serializer(self, instance, **options):
        """
        registers a new serializer or replaces the existing one.

        if `replace=True` is provided. otherwise, it raises an error
        on adding a serializer which is already registered.

        :param AbstractSerializerBase instance: serializer to be registered.
                                                it must be an instance of
                                                AbstractSerializerBase.

        :keyword bool replace: specifies that if there is another registered
                               serializer with the same accepted type,
                               replace it with the new one, otherwise raise an error.
                               defaults to False.

        :raises InvalidSerializerTypeError: invalid serializer type error.
        :raises DuplicatedSerializerError: duplicated serializer error.
        """

        if not isinstance(instance, AbstractSerializerBase):
            raise InvalidSerializerTypeError('Input parameter [{instance}] is '
                                             'not an instance of [{base}].'
                                             .format(instance=instance,
                                                     base=AbstractSerializerBase))

        if instance.accepted_type in self._serializers:
            old_instance = self._serializers.get(instance.accepted_type)
            replace = options.get('replace', False)
            if replace is not True:
                raise DuplicatedSerializerError('There is another registered '
                                                'serializer [{old}] for accepted type '
                                                '[{accepted_type}] but "replace" option '
                                                'is not set, so serializer [{instance}] '
                                                'could not be registered.'
                                                .format(old=old_instance,
                                                        accepted_type=instance.accepted_type,
                                                        instance=instance))

            print_warning('Serializer [{old_instance}] is going '
                          'to be replaced by [{new_instance}].'
                          .format(old_instance=old_instance,
                                  new_instance=instance))

        self._serializers[instance.accepted_type] = instance

    def get_serializers(self, accepted_type):
        """
        gets the registered serializers for given type.

        it returns an empty list if no serializer found for given type.

        :param type accepted_type: gets the serializers which are
                                   registered for the accepted type.

        :rtype: list[AbstractSerializerBase]
        """

        serializers = []
        if accepted_type in self._serializers:
            serializers.append(self._serializers.get(accepted_type))

        for key, instance in self._serializers.items():
            if key is not accepted_type and issubclass(accepted_type, key):
                serializers.append(instance)

        return serializers



# -*- coding: utf-8 -*-
"""
logging masking hooks module.
"""

import pyrin.logging.masking.services as masking_services

from pyrin.logging.decorators import logging_hook
from pyrin.logging.hooks import LoggingHookBase


@logging_hook()
class LoggingHook(LoggingHookBase):
    """
    logging hook class.
    """

    def prepare_data(self, data, **options):
        """
        this method will be called when a log message interpolation is required.

        each subclass must return the modified or same input data.

        :param dict | object data: data that is passed to logging method.

        :returns: modified or same input data.
        :rtype: dict | object
        """

        return masking_services.mask(data, **options)



# -*- coding: utf-8 -*-
"""
caching decorators module.
"""

from functools import update_wrapper

from werkzeug.utils import cached_property as cached_property_base

import pyrin.caching.services as caching_services
import pyrin.utils.function as function_utils

from pyrin.core.decorators import class_property


def cache(*args, **kwargs):
    """
    decorator to register a cache.

    :param object args: cache class constructor arguments.
    :param object kwargs: cache class constructor keyword arguments.

    :keyword bool replace: specifies that if there is another registered
                           cache with the same name, replace it with
                           the new one, otherwise raise an error.
                           defaults to False.

    :keyword int limit: limit count or size of cached items.
                        if not provided, it will be get
                        from `caching` config store.
                        if you want to remove count limit,
                        you could pass `caching.globals.NO_LIMIT`
                        as input. this is only used in complex
                        and some remote caches.

    :keyword int expire: default expire time of cached items in milliseconds.
                         if not provided, it will be get from `caching` config
                         store. this is only used in complex and remote caches.

    :keyword bool refreshable: specifies that cached item's expire time must be
                               extended on each hit. if not provided, will be get
                               from `caching` config store. this is only used in
                               complex caches.

    :keyword bool use_lifo: specifies that items of the cache must
                            be removed in lifo order. if not provided,
                            it will be get from `caching` config store.
                            this is only used in complex caches.

    :keyword int clear_count: number of old items to be removed from cache when
                              the cache is full. if not provided, it will be get
                              from `caching` config store.
                              note that reducing this value to extremely low values
                              will cause a performance issue when the cache becomes full.
                              this is only used in complex caches.

    :keyword bool consider_user: specifies that current user must be included in
                                 key generation. if not provided, it will be get
                                 from `caching` config store.
                                 this is only used in extended, complex and remote
                                 caches.

    :keyword bool persistent: specifies that cached items must be persisted to
                              database on application shutdown, and loaded back
                              on application startup. if not provided, will be
                              get from `caching` config store.
                              this is only used in complex caches.

    :keyword int chunk_size: chunk size to insert values for persistent caches.
                             after each chunk, store will be flushed.
                             if not provided, will be get from `caching` config store.
                             this is only used in complex caches.

    :raises InvalidCacheTypeError: invalid cache type error.
    :raises DuplicatedCacheError: duplicated cache error.

    :returns: cache class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available caches.

        :param type cls: cache class.

        :returns: cache class.
        :rtype: type
        """

        instance = cls(*args, **kwargs)
        caching_services.register_cache(instance, **kwargs)

        return cls

    return decorator


def permanent(*old_method):
    """
    decorator to convert a method or property into a lazy one.

    note that this cache type is permanent and will not consider method inputs
    in caching. the result will be calculated once and then it will be cached.
    each result will be cached using a tuple of class type and method name as
    a key in the cache. so if used on instance or class methods, all instances
    of the same class type will have access to the same shared cache. if you want
    to use a cache per each instance of a class type, you could use `cached_property`
    decorator.

    note that this decorator should only be used on instance or class methods or
    properties. otherwise the result would be inconsistent.
    for stand-alone functions use `permanent_function` decorator.

    :param function | property old_method: the original decorated method or property.

    :returns: method or property result.
    """

    def decorator(method):
        """
        decorates the given method or property and makes it a lazy one.

        :param function | property method: decorated method or property.

        :returns: method or property result.
        """

        def wrapper(self):
            """
            decorates the given method or property and makes it a lazy one.

            :returns: method or property result.
            """

            result = caching_services.try_get('permanent', method, self)
            if result is not None:
                return result

            result = method(self)
            caching_services.try_set('permanent', result, method, self)
            return result

        return update_wrapper(wrapper, method)

    if len(old_method) > 0:
        return decorator(old_method[0])

    return decorator


def permanent_function(*old_func):
    """
    decorator to convert a function into a lazy one.

    note that this cache type is permanent and will not consider function inputs
    in caching. the result will be calculated once and then it will be cached.
    each result will be cached using a tuple of function fully qualified name as
    a key in the cache.

    note that this decorator should only be used on stand-alone functions.
    otherwise the result would be inconsistent. for instance or class methods
    use `permanent` decorator.

    :param function old_func: the original decorated function.

    :returns: function result.
    """

    def decorator(func):
        """
        decorates the given function and makes it a lazy one.

        :param function func: decorated function.

        :returns: function result.
        """

        def wrapper():
            """
            decorates the given function and makes it a lazy one.

            :returns: function result.
            """

            result = caching_services.try_get('permanent', func, None)
            if result is not None:
                return result

            result = func()
            caching_services.try_set('permanent', result, func, None)
            return result

        return update_wrapper(wrapper, func)

    if len(old_func) > 0:
        return decorator(old_func[0])

    return decorator


def extended_permanent(*old_func, **options):
    """
    decorator to convert a method or function into a lazy one.

    note that this cache type is permanent but will consider function inputs
    in caching. the result will be calculated once and then it will be cached.
    each result will be cached using a tuple of parent class type, function fully
    qualified name, inputs, current user and component key as a key in the cache.

    that this decorator could be used on both instance or class level methods and
    properties or stand-alone functions.

    :param function old_func: the original decorated method or function.

    :keyword bool consider_user: specifies that current user must be included in
                                 key generation. if not provided, it will be get
                                 from `caching` config store.

    :returns: function result.
    """

    def decorator(func):
        """
        decorates the given method or function and makes it a lazy one.

        :param function func: decorated method or function.

        :returns: function result.
        """

        def wrapper(*args, **kwargs):
            """
            decorates the given method or function and makes it a lazy one.

            :param object args: function positional arguments.
            :param object kwargs: function keyword arguments.

            :returns: function result.
            """

            result = caching_services.try_get('extended.permanent', func,
                                              args, kwargs, **options)
            if result is not None:
                return result

            result = func(*args, **kwargs)
            caching_services.try_set('extended.permanent', result, func,
                                     args, kwargs, **options)
            return result

        return update_wrapper(wrapper, func)

    if len(old_func) > 0:
        return decorator(old_func[0])

    return decorator


def cached(*old_method, **options):
    """
    decorator to convert a method or function into a lazy one.

    note that this cache type supports expire time and will consider method inputs
    in caching. the result will be calculated once and then it will be cached.
    each result will be cached using a tuple of class type, method name, inputs,
    current user and component key as a key in the cache.

    that this decorator could be used on both instance or class level methods and
    properties or stand-alone functions.

    :param function | property old_method: the original decorated method or function.

    :keyword bool consider_user: specifies that current user must be included in
                                 key generation. if not provided, it will be get
                                 from `caching` config store.

    :keyword int expire: expire time for given key in milliseconds.
                         if not provided, it will be get from `caching`
                         config store.

    :keyword bool refreshable: specifies that cached item's expire time must be
                               extended on each hit. if not provided, it will be
                               get from `caching` config store.

    :returns: method or function result.
    """

    def decorator(method):
        """
        decorates the given method or function and makes it a lazy one.

        :param function | property method: decorated method or function.

        :returns: method or function result.
        """

        def wrapper(*args, **kwargs):
            """
            decorates the given method or function and makes it a lazy one.

            :param object args: function positional arguments.
            :param object kwargs: function keyword arguments.

            :returns: method or function result.
            """

            result = caching_services.try_get('complex', method, args,
                                              kwargs, **options)
            if result is not None:
                return result

            result = method(*args, **kwargs)
            caching_services.try_set('complex', result, method,
                                     args, kwargs, **options)
            return result

        return update_wrapper(wrapper, method)

    if len(old_method) > 0:
        return decorator(old_method[0])

    return decorator


def custom_cached(name, *old_method, **options):
    """
    decorator to convert a method or function into a lazy one.

    note that this cache type supports expire time and will consider method inputs
    in caching. the result will be calculated once and then it will be cached.
    each result will be cached using a tuple of class type, method name, inputs,
    current user and component key as a key in the cache.

    that this decorator could be used on both instance or class level methods and
    properties or stand-alone functions.

    :param str name: the cache name to be used.
                     for example: `redis`, `memcached`, `complex` or ...

    :param function | property old_method: the original decorated method or function.

    :keyword bool consider_user: specifies that current user must be included in
                                 key generation. if not provided, it will be get
                                 from `caching` config store.

    :keyword int expire: expire time for given key in milliseconds.
                         if not provided, it will be get from `caching`
                         config store.

    :keyword bool refreshable: specifies that cached item's expire time must be
                               extended on each hit. if not provided, it will be
                               get from `caching` config store.

    :returns: method or function result.
    """

    def decorator(method):
        """
        decorates the given method or function and makes it a lazy one.

        :param function | property method: decorated method or function.

        :returns: method or function result.
        """

        def wrapper(*args, **kwargs):
            """
            decorates the given method or function and makes it a lazy one.

            :param object args: function positional arguments.
            :param object kwargs: function keyword arguments.

            :returns: method or function result.
            """

            result = caching_services.try_get(name, method, args,
                                              kwargs, **options)
            if result is not None:
                return result

            result = method(*args, **kwargs)
            caching_services.try_set(name, result, method,
                                     args, kwargs, **options)
            return result

        return update_wrapper(wrapper, method)

    if len(old_method) > 0:
        return decorator(old_method[0])

    return decorator


class cached_property(cached_property_base):
    """
    a decorator to convert a property into a cached property.

    the result of the property will be calculated once and cached.
    the cached value is per instance not per type.

    usage example:

    @cached_property
    def is_valid(self):
        return True
    """
    pass


class cached_class_property(class_property):
    """
    a decorator to convert a class property into a cached class property.

    the result of the class property will be calculated once and cached.
    the cached value is per type not per instance.
    it only supports get.

    usage example:

    @cached_class_property
    def is_valid(cls):
        return True
    """

    def __init__(self, method=None):
        """
        initializes an instance of cached_class_property.

        :param function method: decorated method.
        """

        super().__init__(method)
        self.__cache_name__ = self._get_cache_key()

    def __get__(self, instance, cls=None):
        """
        gets the result of decorated method.

        if the result is available in the cache, it will be get from there.
        otherwise the method will be called to produce result.

        :param instance: instance of parent class.
        :param type cls: class type.

        :returns: decorated method result.
        """

        if cls is None:
            cls = type(instance)

        try:
            return vars(cls)[self.__cache_name__]
        except KeyError:
            result = super().__get__(instance, cls)
            setattr(cls, self.__cache_name__, result)
            return result

    def _get_cache_key(self):
        """
        gets the cache key name for current method.

        :rtype: str
        """

        return '__CACHED__{method}__'.format(
            method=function_utils.get_fully_qualified_name(self.fget).upper())



# -*- coding: utf-8 -*-
"""
authorization package.
"""


from pyrin.packaging.base import Package


class AuthorizationPackage(Package):
    """
    authorization package class.
    """

    NAME = __name__
    DEPENDS = []
    COMPONENT_NAME = 'security.authorization.component'



# -*- coding: utf-8 -*-
"""
locale handlers module.
"""

import pyrin.globalization.locale.services as locale_services


locale_services.set_locale_selector(locale_services.get_current_locale)
locale_services.set_timezone_selector(locale_services.get_current_timezone)



# -*- coding: utf-8 -*-
"""
validator auto manager module.
"""

from uuid import UUID
from decimal import Decimal
from datetime import datetime, date, time

import pyrin.validator.services as validator_services
import pyrin.database.model.services as model_services

from pyrin.core.structs import Manager, Context
from pyrin.validator.auto import ValidatorAutoPackage
from pyrin.validator.auto.handler import AutoValidator
from pyrin.validator.handlers.base import ValidatorBase
from pyrin.validator.handlers.dictionary import DictionaryValidator
from pyrin.validator.handlers.uuid import UUIDValidator
from pyrin.validator.handlers.string import StringValidator
from pyrin.validator.handlers.boolean import BooleanValidator
from pyrin.validator.handlers.datetime import DateTimeValidator, DateValidator, TimeValidator, \
    FromDateTimeValidator, ToDateTimeValidator
from pyrin.validator.handlers.number import IntegerValidator, FloatValidator, DecimalValidator
from pyrin.validator.handlers.misc import MaximumValidator, MinimumValidator, \
    RangeValidator, InValidator, NotInValidator


class ValidatorAutoManager(Manager):
    """
    validator auto manager class.
    """

    package_class = ValidatorAutoPackage

    # keywords that will be used to register find validators.
    FROM_KEYWORD = 'from_'
    TO_KEYWORD = 'to_'

    # list of all types that support find range validators.
    FIND_RANGE_VALIDATOR_TYPES = [int, float, Decimal, datetime, date, time]

    def __init__(self):
        """
        initializes an instance of ValidatorAutoManager.
        """

        super().__init__()

        # a dictionary containing python type to validator map.
        # example: {type | str python_type_or_name: type[AbstractValidatorBase] validator_class}
        self._type_to_validator_map = Context()
        self._initialize_map()

    def _initialize_map(self):
        """
        initializes the python type to validator map.
        """

        self._type_to_validator_map[int] = IntegerValidator
        self._type_to_validator_map[float] = FloatValidator
        self._type_to_validator_map[Decimal] = DecimalValidator
        self._type_to_validator_map[bool] = BooleanValidator
        self._type_to_validator_map[str] = StringValidator
        self._type_to_validator_map[dict] = DictionaryValidator
        self._type_to_validator_map[datetime] = DateTimeValidator
        self._type_to_validator_map[date] = DateValidator
        self._type_to_validator_map[time] = TimeValidator
        self._type_to_validator_map[UUID] = UUIDValidator
        self._type_to_validator_map['from_datetime'] = FromDateTimeValidator
        self._type_to_validator_map['to_datetime'] = ToDateTimeValidator
        self._type_to_validator_map['max'] = MaximumValidator
        self._type_to_validator_map['min'] = MinimumValidator
        self._type_to_validator_map['range'] = RangeValidator
        self._type_to_validator_map['in'] = InValidator
        self._type_to_validator_map['not_in'] = NotInValidator

    def _get_validator_class(self, python_type, **options):
        """
        gets the validator class for given python type.

        it return None if no validator found with given type or name.

        :param type | str python_type: python type or name of validator class.

        :keyword bool from_datetime: specifies that if python type is `datetime`,
                                     get the `from_datetime` validator.
                                     if set to False, gets the `to_datetime` validator.
                                     if set to None, gets the `datetime` validator.
                                     defaults to None if not provided.

        :rtype: type[AbstractValidatorBase]
        """

        if python_type is datetime:
            from_datetime = options.get('from_datetime')
            if from_datetime is True:
                python_type = 'from_datetime'
            elif from_datetime is False:
                python_type = 'to_datetime'

        return self._type_to_validator_map.get(python_type)

    def _get_type_validator(self, domain, field, **options):
        """
        gets the required type validator for given field.

        it may return None if `force=False` is given and no validator could be created.

        :param BaseEntity domain: entity type that this field is related to.
        :param InstrumentedAttribute field: field instance.

        :keyword bool for_find: specifies that this validator must only
                                be used on validation for find.
                                defaults to False if not provided.

        :keyword str name: a custom name for this validator.
                           if provided, the name of `field` will be ignored.

        :keyword bool force: specifies that if no validator could be
                             created for given field's type, it must return
                             a base validator for it, otherwise it returns None.
                             defaults to True if not provided.

        :keyword bool from_datetime: specifies that if python type is `datetime`,
                                     get the `from_datetime` validator.
                                     if set to False, gets the `to_datetime` validator.
                                     if set to None, gets the `datetime` validator.
                                     defaults to None if not provided.

        :keyword bool allow_list_for_find: allow value to be list too, on validation
                                           for find. single value is also accepted for
                                           find. if not provided and the field is a
                                           column which has `check_in` set for it,
                                           this will be set to True.

        :rtype: AbstractValidatorBase
        """

        force = options.get('force', True)
        collection_type, python_type = field.get_python_type()
        if collection_type is list:
            options.update(is_list=True)

        validator_class = self._get_validator_class(python_type, **options)
        if validator_class is not None:
            return validator_class(domain, field, **options)

        if force is True:
            return ValidatorBase(domain, field, **options)

        return None

    def _get_range_validator(self, domain, field, **options):
        """
        gets the required range validator for given field.

        it either returns a min validator, max validator, range validator or None.

        :param BaseEntity domain: entity type that this field is related to.
        :param InstrumentedAttribute field: field instance.

        :rtype: AbstractValidatorBase
        """

        validator_class = None
        if field.min_value is not None and field.max_value is not None:
            validator_class = self._get_validator_class('range')

        elif field.min_value is not None:
            validator_class = self._get_validator_class('min')

        elif field.max_value is not None:
            validator_class = self._get_validator_class('max')

        if validator_class is not None:
            return validator_class(domain, field, **options)

        return None

    def _get_in_validator(self, domain, field, **options):
        """
        gets the required in validator for given field.

        it either returns an in validator, not in validator or None.

        :param BaseEntity domain: entity type that this field is related to.
        :param InstrumentedAttribute field: field instance.

        :rtype: AbstractValidatorBase
        """

        validator_class = None
        if field.check_in is not None:
            validator_class = self._get_validator_class('in')

        elif field.check_not_in is not None:
            validator_class = self._get_validator_class('not_in')

        if validator_class is not None:
            return validator_class(domain, field, **options)

        return None

    def _get_find_validator(self, domain, field, **options):
        """
        gets the required find validator for given field.

        it will only be used in find validation and will only validate
        type of value if it is not None.

        :param BaseEntity domain: entity type that this field is related to.
        :param InstrumentedAttribute field: field instance.

        :rtype: AbstractValidatorBase
        """

        find_validator = validator_services.try_get_validator(domain, field.key)
        if find_validator is None:
            find_validator = self._get_type_validator(domain, field, for_find=True)

        return find_validator

    def _get_find_range_validators(self, domain, field, **options):
        """
        gets the required find range validators for given field.

        it returns a tuple of all required find range validators.
        it may return an empty tuple if no find range validator should be created.

        find range validators are constructed with names `from_*` and `to_*`.
        they will only be used in find validation and will only validate
        type of value if it is not None.

        not that for primary key columns, no find range validators will be created.

        :param BaseEntity domain: entity type that this field is related to.
        :param InstrumentedAttribute field: field instance.

        :rtype: tuple
        """

        collection_type, python_type = field.get_python_type()
        if field.primary_key is True or python_type not in self.FIND_RANGE_VALIDATOR_TYPES:
            return tuple()

        from_name = '{prefix}{field}'.format(prefix=self.FROM_KEYWORD, field=field.key)
        from_validator = self._get_type_validator(domain, field, name=from_name,
                                                  for_find=True, from_datetime=True,
                                                  allow_list_for_find=False)

        to_name = '{prefix}{field}'.format(prefix=self.TO_KEYWORD, field=field.key)
        to_validator = self._get_type_validator(domain, field, name=to_name,
                                                for_find=True, from_datetime=False,
                                                allow_list_for_find=False)

        return from_validator, to_validator

    def register_auto_validator(self, domain, field, **options):
        """
        register required auto validator for given field.

        :param BaseEntity domain: entity type that this field is related to.
        :param InstrumentedAttribute field: field instance.
        """

        type_validator = self._get_type_validator(domain, field, **options)
        range_validator = self._get_range_validator(domain, field, **options)
        in_validator = self._get_in_validator(domain, field, **options)
        custom_validator = validator_services.try_get_validator(domain, field.key)
        if custom_validator is not None:
            options.update(replace=True)

        options.update(type_validator=type_validator,
                       range_validator=range_validator,
                       in_validator=in_validator,
                       custom_validator=custom_validator)

        auto_validator = AutoValidator(domain, field, **options)
        validator_services.register_validator(auto_validator, **options)

    def register_find_validator(self, domain, field, **options):
        """
        registers required find validator for given field.

        :param BaseEntity domain: entity type that this field is related to.
        :param InstrumentedAttribute field: field instance.
        """

        find_validator = self._get_find_validator(domain, field, **options)
        options.update(for_find=True)
        validator_services.register_validator(find_validator, **options)

    def register_find_range_validators(self, domain, field, **options):
        """
        registers required find range validators for given field.

        it returns the count of registered find range validators.

        :param BaseEntity domain: entity type that this field is related to.
        :param InstrumentedAttribute field: field instance.

        :rtype: int
        """

        registered = 0
        find_range_validators = self._get_find_range_validators(domain, field, **options)
        options.update(for_find=True)
        for item in find_range_validators:
            validator_services.register_validator(item, **options)
            registered += 1

        return registered

    def register_auto_validators(self):
        """
        registers all auto validators.

        it returns the count of registered auto validators.

        :rtype: int
        """

        entities = model_services.get_entities()
        registered = 0
        for entity in entities:
            for column in entity.all_instrumented_attributes:
                if column.validated is True:
                    self.register_auto_validator(entity, column)
                    registered += 1

                if column.validated_find is True:
                    self.register_find_validator(entity, column)
                    registered += 1

                if column.validated_range is True:
                    count = self.register_find_range_validators(entity, column)
                    registered += count

        return registered



# -*- coding: utf-8 -*-
"""
cli base module.
"""

import subprocess

from subprocess import CalledProcessError
from threading import Lock
from abc import abstractmethod

from pyrin.cli.mixin.param import CLIParamMixin
from pyrin.core.structs import CoreObject, MultiSingletonMeta
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.cli.exceptions import InvalidCLIHandlerNameError


class CLIHandlerSingletonMeta(MultiSingletonMeta):
    """
    cli handler singleton meta class.

    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class AbstractCLIHandlerBase(CoreObject, metaclass=CLIHandlerSingletonMeta):
    """
    abstract cli handler base class.
    """

    @abstractmethod
    def get_name(self):
        """
        gets current handler name, the handler will be registered with this name.

        the name must be the exact command name that this handler handles.
        for example `revision`.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def execute(self, **options):
        """
        executes the current cli command with given inputs.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: execution result.
        """

        raise CoreNotImplementedError()


class CLIHandlerBase(CLIParamMixin, AbstractCLIHandlerBase):
    """
    cli handler base class.

    all application cli handlers must be subclassed from this.
    """

    def __init__(self, name, **options):
        """
        initializes an instance of CLIHandlerBase.

        :param str name: the handler name that should be registered
                         with. this name must be the exact name that
                         this handler must emmit to cli.

        :keyword bool help_param: specifies that help param must
                                  be added for this handler.
                                  defaults to True if not provided.

        :raises InvalidCLIHandlerNameError: invalid cli handler name error.
        """

        if name in (None, '') or name.isspace():
            raise InvalidCLIHandlerNameError('CLI handler name must be provided.')

        self._name = name
        super().__init__(**options)

    def __str__(self):
        """
        gets the string representation of current cli handler.

        :rtype: str
        """

        return '{base} [{handler}]'.format(base=super().__str__(), handler=self._name)

    def get_name(self):
        """
        gets current handler name, the handler will be registered with this name.

        the name must be the exact command name that this handler handles.
        for example `revision`.

        :rtype: str
        """

        return self._name

    def execute(self, **options):
        """
        executes the current cli command with given inputs.

        :rtype: int
        """

        return self._execute(**options)

    def _execute(self, **options):
        """
        executes the current cli command with given inputs.

        :rtype: int
        """

        commands = []
        self._inject_command_name(commands)
        processed_inputs = self._process_inputs(**options)
        self._bind_cli_arguments(commands, **processed_inputs)
        common_options = self._get_common_cli_options()
        common_options.extend(commands)
        return self._execute_on_cli(common_options)

    def _inject_command_name(self, commands):
        """
        injects the command name into given commands list.

        this method could be overridden in subclasses if a
        different behaviour is required.

        :param list commands: a list of all commands to inject
                              this command's name into it.
        """

        commands.insert(0, self.get_name())

    def _execute_on_cli(self, commands):
        """
        executes the current cli command with given inputs.

        :param list commands: a list of all commands and their
                              values to be sent to cli command.

        :rtype: int
        """

        try:
            result = subprocess.check_call(commands)
            return self._process_return_value(result)
        except CalledProcessError as error:
            return self._process_return_value(error.returncode)

    def _process_return_value(self, result):
        """
        processes the return value from cli command execution.

        this method returns the None value and is intended to be overridden
        by subclasses if required. if a cli handlers group does need to return
        a value, it could override this method and return the desired value.

        :param int result: result value returned from cli command.

        :rtype: object
        """

        return None

    def _get_common_cli_options(self):
        """
        gets the list of common cli options.

        this method is intended to be overridden by subclasses
        to inject some common cli options into the commands list.
        subclasses must return a list containing all common options.
        the common options will be inserted into beginning of the commands list.

        :rtype: list
        """

        return []



# -*- coding: utf-8 -*-
"""
cli params module.
"""

from pyrin.cli.arguments import ArgumentBase, BooleanArgument


class CLIParamBase(ArgumentBase):
    """
    cli param base class.

    all cli param classes must be subclassed from this.
    """

    def process_inputs(self, **options):
        """
        processes the inputs in given dict.

        :rtype: dict
        """

        return self._process_inputs(**options)

    def _process_inputs(self, **options):
        """
        processes the inputs in given dict.

        subclasses could override this method and modify inputs as needed.
        if the modification is a static one, you could provide default in
        the param initialization, but if the modification is not static,
        for example, a callable must be used to generate value based on
        current situation, you must override this method to modify inputs and
        inject the correct value. subclasses must call `super()._process_inputs(**options)`
        after their work has be done.

        :rtype: dict
        """

        return options


class HelpParam(BooleanArgument, CLIParamBase):
    """
    help param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of HelpParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to None if not provided.
        """

        super().__init__('help', '--help', default=default)


class VerboseParam(BooleanArgument, CLIParamBase):
    """
    verbose param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of VerboseParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to None if not provided.
        """

        super().__init__('verbose', '--verbose', default=default)



# -*- coding: utf-8 -*-
"""
api manager module.
"""

import pyrin.logging.services as logging_services
import pyrin.configuration.services as config_services
import pyrin.processor.response.services as response_services
import pyrin.security.session.services as session_services

from pyrin.api import APIPackage
from pyrin.api.exceptions import InvalidAPIHookTypeError
from pyrin.api.hooks import APIHookBase
from pyrin.core.mixin import HookMixin
from pyrin.core.structs import Manager
from pyrin.core.enumerations import ServerErrorResponseCodeEnum
from pyrin.core.globals import _


class APIManager(Manager, HookMixin):
    """
    api manager class.
    """

    package_class = APIPackage
    LOGGER = logging_services.get_logger('api')
    hook_type = APIHookBase
    invalid_hook_type_error = InvalidAPIHookTypeError

    def handle_http_error(self, exception):
        """
        handles http exceptions.

        note that normally you should never call this method manually.

        :param HTTPException exception: exception instance.

        :returns: tuple[dict | object, int, CoreHeaders]
        :rtype: tuple
        """

        self._log_exception(exception)
        return response_services.make_exception_response(exception,
                                                         request_id=self._get_request_id())

    def handle_server_business_error(self, exception):
        """
        handles server internal core business exceptions.

        note that normally you should never call this method manually.

        :param CoreBusinessException exception: core business exception instance.

        :returns: tuple[dict | object, int, CoreHeaders]
        :rtype: tuple
        """

        self._log_exception(exception)
        return response_services.make_exception_response(exception,
                                                         request_id=self._get_request_id(),
                                                         data=exception.data)

    def handle_server_error(self, exception):
        """
        handles server internal core exceptions.

        note that normally you should never call this method manually.
        in any environment which debug mode is False, the original error
        message will be replaced by a generic error message before being
        sent to client for security reasons.

        :param CoreException exception: core exception instance.

        :returns: tuple[dict | object, int, CoreHeaders]
        :rtype: tuple
        """

        self._log_exception(exception)
        if config_services.get_active('environment', 'debug') is True:
            return response_services.make_exception_response(exception,
                                                             request_id=self._get_request_id(),
                                                             data=exception.data)

        return response_services.make_error_response(self._get_generic_error_message(),
                                                     code=exception.code,
                                                     request_id=self._get_request_id())

    def handle_server_unknown_error(self, exception):
        """
        handles unknown server internal exceptions.

        note that normally you should never call this method manually.
        in any environment which debug mode is False, the original error
        message will be replaced by a generic error message before being
        sent to client for security reasons.

        :param Exception exception: exception instance.

        :returns: tuple[dict | object, int, CoreHeaders]
        :rtype: tuple
        """

        self._log_exception(exception)
        if config_services.get_active('environment', 'debug') is True:
            return response_services.make_exception_response(exception,
                                                             code=ServerErrorResponseCodeEnum.
                                                             INTERNAL_SERVER_ERROR,
                                                             request_id=self._get_request_id())

        return response_services.make_error_response(self._get_generic_error_message(),
                                                     code=ServerErrorResponseCodeEnum.
                                                     INTERNAL_SERVER_ERROR,
                                                     request_id=self._get_request_id())

    def _get_generic_error_message(self):
        """
        gets the generic error message to be sent to client.

        for any environment which debug mode is False.

        :rtype: str
        """

        return _('Application has been encountered an error. Please '
                 'contact the support team if problem persists.')

    def _get_request_id(self):
        """
        gets current request id.

        :rtype: uuid.UUID
        """

        return session_services.get_current_request_id()

    def _log_exception(self, exception):
        """
        logs the input exception.

        :param Exception exception: exception instance to be logged.
        """

        self.LOGGER.exception(str(exception))
        self._exception_occurred(exception)

    def _exception_occurred(self, error, **options):
        """
        this method will call `exception_occurred` method on all registered hooks.

        :param Exception error: exception instance that has been occurred.
        """

        for hook in self._get_hooks():
            hook.exception_occurred(error, **options)



# -*- coding: utf-8 -*-
"""
deserializer handlers package.
"""

from pyrin.packaging.base import Package


class DeserializerHandlersPackage(Package):
    """
    deserializer handlers package class.
    """

    NAME = __name__
    DEPENDS = []



# -*- coding: utf-8 -*-
"""
utils dictionary module.
"""

from collections import OrderedDict

import pyrin.utils.string as string_utils

from pyrin.core.structs import DTO
from pyrin.utils.exceptions import InputNotCallableError, DictKeyPrefixIsNotProvidedError


def change_key_case(value, converter, **options):
    """
    returns a copy of input dict with all it's keys
    and nested keys cases modified using given converter.
    note that this only considers direct nested dicts, if there is a
    list that contains a dict as value, this method won't change that.

    :param dict value: dict to change it's keys cases.

    :param callable converter: a callable to use as case converter.
                               it should be a callable with a signature
                               similar to below example.
                               case_converter(input_dict).

    :raises InputNotCallableError: input not callable error.

    :rtype: dict
    """

    if not callable(converter):
        raise InputNotCallableError('Input parameter [{converter}] is not callable.'
                                    .format(converter=str(converter)))

    result_dict = DTO()
    for key, val in value.items():
        if isinstance(val, dict):
            val = change_key_case(val, converter, **options)
        result_dict[converter(key)] = val
    return result_dict


def make_key_upper(value):
    """
    returns a copy of input dict with all it's
    keys and nested keys cases made upper.
    note that this only considers direct nested dicts, if there is a
    list that contains a dict as value, this method won't change that.

    :param dict value: dict to make it's keys uppercase.

    :rtype: dict
    """

    return change_key_case(value, string_utils.upper)


def make_key_lower(value):
    """
    returns a copy of input dict with all it's
    keys and nested keys cases made lower.
    note that this only considers direct nested dicts, if there is a
    list that contains a dict as value, this method won't change that.

    :param dict value: dict to make it's keys lowercase.

    :rtype: dict
    """

    return change_key_case(value, string_utils.lower)


def remove_keys(value, prefix):
    """
    removes all keys from given dict which starting with given prefix and
    also those keys that their names are exactly like prefixed ones, but without prefix.
    for example: dict(name='some_name', value='value', __value='another_value')
    if prefix is double underscores `__` then this function will return a new dict
    result = dict(name='some_name') with just the name key.
    note that this function just loops through top level keys.

    :param dict value: dict to remove keys from it.
    :param str prefix: keys prefix that should be removed from dict.

    :raises DictKeyPrefixIsNotProvidedError: dict key prefix is not provided error.

    :rtype: dict
    """

    if prefix in (None, '') or prefix.isspace():
        raise DictKeyPrefixIsNotProvidedError('Dictionary key prefix must be provided '
                                              'to remove keys with such prefix.')

    prefix_length = len(prefix)
    prefixed_to_remove = [key for key in value.keys() if key.startswith(prefix)]
    original_to_remove = [key[prefix_length:] for key in prefixed_to_remove]
    prefixed_to_remove.extend(original_to_remove)
    result = DTO(**value)
    for key in prefixed_to_remove:
        result.pop(key, None)

    return result


def sort_by_value(value, reverse=False):
    """
    sorts a dictionary by its values.

    :param dict value: dict to be sorted.

    :param bool reverse: sort by descending order.
                         defaults to False if not provided.

    :rtype: OrderedDict
    """

    result = sorted(value.items(), key=lambda x: x[1], reverse=reverse)
    return OrderedDict(result)


def sort_by_key(value, reverse=False):
    """
    sorts a dictionary by its keys.

    :param dict value: dict to be sorted.

    :param bool reverse: sort by descending order.
                         defaults to False if not provided.

    :rtype: OrderedDict
    """

    result = sorted(value.items(), key=lambda x: x[0], reverse=reverse)
    return OrderedDict(result)


def sort_by_key_length(value, reverse=False):
    """
    sorts a dictionary by its keys length.

    note that keys must be strings.

    :param dict value: dict to be sorted.

    :param bool reverse: sort by descending order.
                         defaults to False if not provided.

    :rtype: OrderedDict
    """

    result = sorted(value.items(), key=lambda x: len(x[0]), reverse=reverse)
    return OrderedDict(result)



# -*- coding: utf-8 -*-
"""
security services module.
"""

import pyrin.configuration.services as config_services

from pyrin.application.services import get_component
from pyrin.caching.decorators import custom_cached
from pyrin.security import SecurityPackage


def get_password_hash(password, **options):
    """
    gets the given password's hash.

    :param str password: password to get it's hash.

    :keyword bool is_encrypted: specifies that given password is encrypted.
                                defaults to False if not provided.

    :raises InvalidPasswordLengthError: invalid password length error.

    :rtype: str
    """

    return get_component(SecurityPackage.COMPONENT_NAME).get_password_hash(password, **options)


def encrypt(text, **options):
    """
    encrypts the given text and returns the encrypted value.

    :param str text: text to be encrypted.

    :raises InvalidEncryptionTextLengthError: invalid encryption text length error.

    :rtype: str
    """

    return get_component(SecurityPackage.COMPONENT_NAME).encrypt(text, **options)


cache_name = config_services.get('security', 'permission', 'cache_name')
cache_expire = config_services.get('security', 'permission', 'cache_expire')
if cache_name is not None:
    @custom_cached(cache_name, expire=cache_expire, consider_user=True, refreshable=False)
    def has_permission(user, permissions, **options):
        """
        gets a value indicating that given user has the specified permissions.

        :param user: user identity to check its permissions.
        :param list[PermissionBase] permissions: permissions to check for user.

        :rtype: bool
        """

        return get_component(SecurityPackage.COMPONENT_NAME).has_permission(user, permissions,
                                                                            **options)
else:
    def has_permission(user, permissions, **options):
        """
        gets a value indicating that given user has the specified permissions.

        :param user: user identity to check its permissions.
        :param list[PermissionBase] permissions: permissions to check for user.

        :rtype: bool
        """

        return get_component(SecurityPackage.COMPONENT_NAME).has_permission(user, permissions,
                                                                            **options)



# -*- coding: utf-8 -*-
"""
validator auto hooks module.
"""

import pyrin.validator.auto.services as validator_auto_services

from pyrin.database.model.decorators import model_hook
from pyrin.database.model.hooks import ModelHookBase
from pyrin.utils.custom_print import print_info


@model_hook()
class ModelHook(ModelHookBase):
    """
    model hook class.
    """

    def after_entities_collected(self):
        """
        this method will be called after all application entities have been collected.
        """

        count = validator_auto_services.register_auto_validators()
        if count > 0:
            print_info('Total of [{count}] auto validators registered.'.format(count=count))



# -*- coding: utf-8 -*-
"""
utils sqlalchemy module.
"""

from sqlalchemy.sql import quoted_name
from sqlalchemy.engine import result_tuple
from sqlalchemy import inspect as sqla_inspect, Table, text, asc, desc, CheckConstraint

import pyrin.utils.datetime as datetime_utils
import pyrin.utils.string as string_utils

from pyrin.core.globals import _
from pyrin.core.globals import LIST_TYPES
from pyrin.utils.exceptions import InvalidRowResultFieldsAndValuesError, \
    FieldsAndValuesCountMismatchError, CheckConstraintValuesRequiredError, \
    MultipleDeclarativeClassesFoundError, InvalidOrderingColumnError


LIKE_CHAR_COUNT_LIMIT = 20


def like_both(value, start='%', end='%'):
    """
    gets a copy of string with `%` or couple of `_` values attached to both ends.

    it is to be used in like operator.

    :param str value: value to be processed.

    :param str start: start place holder to be prefixed.
                      it could be `%` or couple of `_` values
                      for exact matching.
                      defaults to `%` if not provided.

    :param str end: end place holder to be appended.
                    it could be `%` or couple of `_` values
                    for exact matching.
                    defaults to `%` if not provided.

    :rtype: str
    """

    if value is None:
        return None

    value = like_prefix(value, start)
    value = like_suffix(value, end)

    return value


def like_prefix(value, start='%'):
    """
    gets a copy of string with `%` or couple of `_` values attached to beginning.

    it is to be used in like operator.

    :param str value: value to be processed.

    :param str start: start place holder to be prefixed.
                      it could be `%` or couple of `_` values
                      for exact matching.
                      defaults to `%` if not provided.

    :rtype: str
    """

    if value is None:
        return None

    return '{start}{value}'.format(start=start, value=value)


def like_suffix(value, end='%'):
    """
    gets a copy of string with `%` or couple of `_` values attached to end.

    it is to be used in like operator.

    :param str value: value to be processed.

    :param str end: end place holder to be appended.
                    it could be `%` or couple of `_` values
                    for exact matching.
                    defaults to `%` if not provided.

    :rtype: str
    """

    if value is None:
        return None

    return '{value}{end}'.format(value=value, end=end)


def _process_place_holder(value, count):
    """
    processes the value and generates a place holder with count of `_` chars.

    this value could be used in like operator.

    :param str value: value to be processed.
    :param int count: count of `_` chars to be attached.

    :note count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`, if
                 the provided value goes upper than this limit, a
                 `%` will be attached instead of it. this limit is
                 for security reason.

    :rtype: str
    """

    if value is None:
        return ''

    if count is None or count <= 0:
        return ''

    place_holder = None
    if count > LIKE_CHAR_COUNT_LIMIT:
        place_holder = '%'

    if place_holder is None:
        place_holder = '_' * count

    return place_holder


def like_exact_both(value, count):
    """
    gets a copy of string with `count` number of `_` attached to both ends.

    it is to be used in like operator.

    :param str value: value to be processed.
    :param int count: count of `_` chars to be attached.

    :note count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`, if
                 the provided value goes upper than this limit, a
                 `%` will be attached instead of it. this limit is
                 for security reason.

    :rtype: str
    """

    place_holder = _process_place_holder(value, count)
    return like_both(value, place_holder, place_holder)


def like_exact_prefix(value, count):
    """
    gets a copy of string with `count` number of `_` attached to beginning.

    it is to be used in like operator.

    :param str value: value to be processed.
    :param int count: count of `_` chars to be attached.

    :note count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`, if
                 the provided value goes upper than this limit, a
                 `%` will be attached instead of it. this limit is
                 for security reason.

    :rtype: str
    """

    place_holder = _process_place_holder(value, count)
    return like_prefix(value, place_holder)


def like_exact_suffix(value, count):
    """
    gets a copy of string with `count` number of `_` attached to end.

    it is to be used in like operator.

    :param str value: value to be processed.
    :param int count: count of `_` chars to be attached.

    :note count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`, if
                 the provided value goes upper than this limit, a
                 `%` will be attached instead of it. this limit is
                 for security reason.

    :rtype: str
    """

    place_holder = _process_place_holder(value, count)
    return like_suffix(value, place_holder)


def add_range_clause(clauses, column, value_lower, value_upper,
                     include_equal_to_lower=True,
                     include_equal_to_upper=True,
                     **options):
    """
    adds range comparison into given clauses using specified inputs.

    :param list clauses: clause list to add range clause to it.
    :param CoreColumn column: entity column to add range clause for it.
    :param object value_lower: lower bound of range clause.
    :param object value_upper: upper bound of range clause.

    :param include_equal_to_lower: specifies that lower value
                                   should be considered in range.
                                   defaults to True if not provided.

    :param include_equal_to_upper: specifies that upper value
                                   should be considered in range.
                                   defaults to True if not provided.
    """

    if value_lower is None and value_upper is None:
        return

    if value_lower is not None and value_lower == value_upper:
        clauses.append(column == value_lower)
    else:
        if value_lower is not None:
            if include_equal_to_lower is True:
                clauses.append(column >= value_lower)
            else:
                clauses.append(column > value_lower)
        if value_upper is not None:
            if include_equal_to_upper is True:
                clauses.append(column <= value_upper)
            else:
                clauses.append(column < value_upper)


def add_datetime_range_clause(clauses, column,
                              value_lower, value_upper,
                              include_equal_to_lower=True,
                              include_equal_to_upper=True,
                              **options):
    """
    adds datetime range comparison into given clauses using specified inputs.

    if the inputs are date objects, they will be converted to datetime with client
    timezone and `consider_begin_of_day` and `consider_end_of_day` will also
    considered as True.

    :param list clauses: clause list to add datetime range clause to it.
    :param CoreColumn column: entity column to add datetime range clause for it.
    :param datetime | date value_lower: lower bound of datetime range clause.
    :param datetime | date value_upper: upper bound of datetime range clause.

    :param include_equal_to_lower: specifies that lower datetime
                                   should be considered in range.
                                   defaults to True if not provided.

    :param include_equal_to_upper: specifies that upper datetime
                                   should be considered in range.
                                   defaults to True if not provided.

    :keyword bool consider_begin_of_day: specifies that consider begin
                                         of day for lower datetime.
                                         defaults to False if not provided.

    :keyword bool consider_end_of_day: specifies that consider end
                                       of day for upper datetime.
                                       defaults to False if not provided.
    """

    value_lower, value_upper = datetime_utils.normalize_datetime_range(value_lower,
                                                                       value_upper,
                                                                       **options)

    add_range_clause(clauses, column,
                     value_lower, value_upper,
                     include_equal_to_lower,
                     include_equal_to_upper,
                     **options)


def add_comparison_clause(clauses, column, value, **options):
    """
    adds list or single comparison into clauses based on given value.

    if the value type is any of list, tuple or set, it generates an
    `in()` comparison, otherwise it generates a simple `==` comparison.

    :param list clauses: clause list to add comparison clause to it.
    :param CoreColumn column: entity column to add comparison clause for it.
    :param object | list[object] value: value to add comparison for it.
    """

    if value is not None:
        if isinstance(value, LIST_TYPES):
            clauses.append(column.in_(value))
        else:
            clauses.append(column == value)


def create_row_result(fields, values):
    """
    creates a row result object with given fields and values.

    this object type is returned by sqlalchemy `Query` when there
    is column names or multiple entities in query.

    :param list[str] fields: field names of the result object.

    :param list[object] values: values to be mapped to fields.
                                they must be in the same order as fields.

    :raises InvalidRowResultFieldsAndValuesError: invalid row result fields
                                                  and values error.

    :raises FieldsAndValuesCountMismatchError: fields and values count mismatch error.

    :rtype: ROW_RESULT
    """

    if fields is None or values is None:
        raise InvalidRowResultFieldsAndValuesError('Input parameters "fields" and '
                                                   '"values" must both be provided, '
                                                   'they could not be None.')

    if len(fields) != len(values):
        raise FieldsAndValuesCountMismatchError('The length of "fields" which is '
                                                '[{fields}] and "values" which is '
                                                '[{values}] does not match.'
                                                .format(fields=len(fields),
                                                        values=len(values)))
    result = result_tuple(fields)
    return result(values)


def check_constraint(column, values, **options):
    """
    generates a check constraint for given column and values.

    by default, it generates an `in` check, but this could be changed to
    `not in` by providing `use_in=False` in options.

    if the first item of values is a string, all values will be quoted.

    :param str column: column name to be used in check constraint.
    :param list values: values to be used in check constraint.

    :keyword bool use_in: specifies that it must generate an `in` check.
                          otherwise it generates a `not in` check.
                          defaults to True if not provided.

    :keyword **options: all other keyword arguments will be passed to
                        underlying `CheckConstraint` constructor.

    :raises CheckConstraintValuesRequiredError: check constraint values required error.

    :rtype: CheckConstraint
    """

    if values is None or len(values) <= 0:
        raise CheckConstraintValuesRequiredError('Values for generating a check '
                                                 'constraint must be provided.')

    converter = str
    is_string = isinstance(values[0], str)
    if is_string:
        converter = string_utils.quote
    use_in = options.pop('use_in', True)
    condition = 'in'
    if use_in is False:
        condition = 'not in'

    string_values = ', '.join(converter(item) for item in values)
    sql_text = '{column} {condition} ({values})'.format(column=quoted_name(column, True),
                                                        condition=condition,
                                                        values=string_values)
    options.update(sqltext=sql_text)
    return CheckConstraint(**options)


def range_check_constraint(column, min_value=None, max_value=None, **options):
    """
    generates a range check constraint for given column and values.

    if the values are string, they will be quoted.
    if only one value is provided, a max or min constraint will be generated.

    :param str column: column name to be used in check constraint.
    :param object min_value: min value to be used in check constraint.
    :param object max_value: max value to be used in check constraint.

    :keyword **options: all other keyword arguments will be passed to
                        underlying `CheckConstraint` constructor.

    :raises CheckConstraintValuesRequiredError: check constraint values required error.

    :rtype: CheckConstraint
    """

    if min_value is None and max_value is None:
        raise CheckConstraintValuesRequiredError('Values for generating a range check '
                                                 'constraint must be provided.')

    sql_text = None
    if min_value is not None and max_value is not None:
        sql_text = '{column} >= {min} and {column} <= {max}'
    elif min_value is not None:
        sql_text = '{column} >= {min}'
    else:
        sql_text = '{column} <= {max}'

    if isinstance(min_value, str):
        min_value = string_utils.quote(min_value)

    if isinstance(max_value, str):
        max_value = string_utils.quote(max_value)

    sql_text = sql_text.format(column=quoted_name(column, True),
                               min=min_value, max=max_value)

    options.update(sqltext=sql_text)
    return CheckConstraint(**options)


def get_class_by_table(base, table, **options):
    """
    gets declarative class associated with given table.

    if no class is found this function returns `None`.
    if multiple classes were found (polymorphic cases) additional `data` parameter
    can be given to hint which class to return.

    :param type[BaseEntity] base: declarative base model.
    :param Table table: sqlalchemy table object.

    :keyword dict data: data row to determine the class in polymorphic scenarios.

    :keyword bool raise_multi: specifies that if multiple classes found and
                               also provided data could not help, raise an error.
                               otherwise return None.
                               defaults to True if not provided.

    :note: this code is taken from sqlalchemy-utils project.
    https://github.com/kvesteri/sqlalchemy-utils

    for example:

    class User(CoreEntity):
        _table = 'entity'
        id = AutoPKColumn()
        name = StringColumn()

    get_class_by_table(CoreEntity, User.__table__) -> User class

    this function also supports models using single table inheritance.
    additional data parameter should be provided in these cases.

    for example:

    class Entity(CoreEntity):
        _table = 'entity'
        id = AutoPKColumn()
        name = StringColumn()
        type = StringColumn()
        __mapper_args__ = {
            'polymorphic_on': type,
            'polymorphic_identity': 'entity'
        }

    class User(Entity):
        __mapper_args__ = {
            'polymorphic_identity': 'user'
        }

    get_class_by_table(CoreEntity, Entity.__table__, {'type': 'entity'}) -> Entity class
    get_class_by_table(CoreEntity, Entity.__table__, {'type': 'user'}) -> User class

    it also supports extended entities with unlimited depth, it returns the correct child
    entity.

    for example:

    class EntityBase(CoreEntity):
        _table = 'entity'
        id = AutoPKColumn()

    class Entity2(EntityBase):
        _extend_existing = True
        name = StringColumn()

    class Entity3(Entity2):
        _extend_existing = True
        age = CoreColumn(Integer)

    get_class_by_table(CoreEntity, EntityBase.__table__) -> Entity3 class

    :raises MultipleDeclarativeClassesFoundError: multiple declarative classes found error.

    :returns: declarative class or None.
    :rtype: type[BaseEntity]
    """

    data = options.get('data')
    raise_multi = options.get('raise_multi', True)
    found_classes = []
    for item in base.registry.mappers:
        if len(item.tables) > 0 and item.tables[0] is table:
            found_classes.append(item.entity)

    current_count = len(found_classes)
    temp = list(found_classes)
    if current_count > 1:
        for item in found_classes:
            if item.extend_existing is not True:
                temp.remove(item)

    if current_count > len(temp) > 0:
        found_classes = temp
        if len(found_classes) > 1:
            hierarchies = []
            for item in found_classes:
                hierarchies.append(set(item.__mro__))

            hierarchies = sorted(hierarchies, key=len, reverse=True)
            last = hierarchies[0]
            others = hierarchies[1:]
            union = set().union(*others)
            result = union.symmetric_difference(last)
            if len(result) == 1:
                return result.pop()

        elif len(found_classes) == 1:
            return found_classes[0]

    found_classes = set(found_classes)
    if len(found_classes) > 1:
        if not data:
            if raise_multi is True:
                raise MultipleDeclarativeClassesFoundError('Multiple declarative classes found '
                                                           'for table [{table}]. please provide '
                                                           '"data" parameter for this function '
                                                           'to be able to determine polymorphic '
                                                           'scenarios.'.format(table=table.name))
        else:
            for cls in found_classes:
                mapper = sqla_inspect(cls)
                polymorphic_on = mapper.polymorphic_on.name
                if polymorphic_on in data:
                    if data[polymorphic_on] == mapper.polymorphic_identity:
                        return cls

            if raise_multi is True:
                raise MultipleDeclarativeClassesFoundError('Multiple declarative classes found '
                                                           'for table [{table}]. given data row '
                                                           'does not match any polymorphic '
                                                           'identity of the found classes.'
                                                           .format(table=table.name))
    elif found_classes:
        return found_classes.pop()

    return None


def is_valid_column_name(column):
    """
    gets a value indicating that given column name is valid.

    :param str column: column name.

    :rtype: bool
    """

    if not isinstance(column, str) or len(column) <= 0:
        return False

    column = column.replace('+', '').replace('-', '').strip()
    return len(column) > 0 and ' ' not in column


def get_ordering_info(column):
    """
    gets a tuple containing ordering info for given column name.

    it returns a tuple of two item, first item is column name and
    second item is ordering type which is an `UnaryExpression` from `asc` or `desc`.

    default ordering is ascending, but it could be changed to descending
    by prefixing `-` to column name.

    for example:

    age -> ordering for age column ascending.
    -age -> ordering for age column descending.

    :param str column: column name to get its ordering info.

    :rtype: tuple[str, UnaryExpression]
    """

    order_type = asc
    if column.startswith(('-', '+')):
        if column[0] == '-':
            order_type = desc
        column = column[1:]

    return column, order_type


def get_ordering_criterion(*columns, valid_columns=None, ignore_invalid=True):
    """
    gets required criterion for given columns ordering.

    default ordering is ascending, but it could be changed to descending
    by prefixing `-` to column names.

    this method always ignores empty strings and None values.

    for example:

    name, +age -> ordering for name and age columns both ascending.
    name, -age -> ordering for name ascending and age descending.

    :param str columns: column names to get their ordering criterion.
                        they must be the actual table column names.

    :param list[str] valid_columns: valid column names for ordering.
                                    defaults to None if not provided.
                                    they must be the actual table column
                                    names.

    :param bool ignore_invalid: specifies that if provided columns are
                                not in valid column names, ignore them
                                instead of raising an error. note that
                                this only has effect if `valid_columns`
                                is provided. defaults to True.

    :raises InvalidOrderingColumnError: invalid ordering column error.

    :rtype: tuple[UnaryExpression]
    """

    result = []
    for item in columns:
        if is_valid_column_name(item):
            name, order_type = get_ordering_info(item)
            if valid_columns is None or name in valid_columns:
                result.append(order_type(text(name)))
            elif valid_columns is not None and name not in valid_columns:
                if ignore_invalid is False:
                    raise InvalidOrderingColumnError(_('Column [{name}] is not valid '
                                                       'for ordering.').format(name=name))

    return tuple(result)



# -*- coding: utf-8 -*-
"""
permission hooks module.
"""

import pyrin.security.permission.services as permission_services
import pyrin.configuration.services as configs_services

from pyrin.application.decorators import application_hook
from pyrin.application.hooks import ApplicationHookBase
from pyrin.utils.custom_print import print_info


@application_hook()
class ApplicationHook(ApplicationHookBase):
    """
    application hook class.
    """

    def prepare_runtime_data(self):
        """
        this method will be get called after application has been fully initialized.

        any changes of this method to database, will be committed automatically by
        application. so you should not commit anything in the hook. if you do commit
        manually, unexpected behaviors may occur.

        note that this method will not get called when application starts in scripting mode.
        """

        synchronize = configs_services.get('security', 'permission', 'synchronize')
        if synchronize is True:
            print_info('Synchronizing permissions...')
            permission_services.synchronize_all()



# -*- coding: utf-8 -*-
"""
mimetype manager module.
"""

from pyrin.core.structs import Manager, Context
from pyrin.processor.mimetype import MIMETypePackage
from pyrin.processor.mimetype.interface import AbstractMIMETypeHandlerBase
from pyrin.utils.custom_print import print_warning
from pyrin.processor.mimetype.exceptions import InvalidMIMETypeHandlerTypeError, \
    DuplicatedMIMETypeHandlerError


class MIMETypeManager(Manager):
    """
    mimetype manager class.
    """

    package_class = MIMETypePackage

    def __init__(self):
        """
        initializes an instance of MIMETypeManager.
        """

        super().__init__()

        # a dictionary containing information of registered mimetype handlers.
        # example: dict(type accepted_type: list[AbstractMIMETypeHandlerBase] instances)
        self._handlers = Context()

    def get_mimetype(self, value, **options):
        """
        gets the mimetype of given value.

        returns None if it could not detect the correct mimetype.

        :param object value: value to detect its mimetype.

        :returns: mimetype name
        :rtype: str
        """

        options.update(accepted_type=type(value))
        for handler in self.get_mimetype_handlers(**options):
            mimetype = handler.mimetype(value, **options)
            if mimetype is not None:
                return mimetype
            continue

        return None

    def register_mimetype_handler(self, instance, **options):
        """
        registers a new mimetype handler or replaces the existing one.

        if `replace=True` is provided. otherwise, it raises an error
        on adding a mimetype handler which is already registered.

        :param AbstractMIMETypeHandlerBase instance: mimetype handler to be registered.
                                                     it must be an instance of
                                                     AbstractMIMETypeHandlerBase.

        :keyword bool replace: specifies that if there is another registered
                               mimetype handler with the same name and accepted type,
                               replace it with the new one, otherwise raise an error.
                               defaults to False.

        :raises InvalidMIMETypeHandlerTypeError: invalid mimetype handler type error.
        :raises DuplicatedMIMETypeHandlerError: duplicated mimetype handler error.
        """

        if not isinstance(instance, AbstractMIMETypeHandlerBase):
            raise InvalidMIMETypeHandlerTypeError('Input parameter [{instance}] is '
                                                  'not an instance of [{base}].'
                                                  .format(instance=instance,
                                                          base=InvalidMIMETypeHandlerTypeError))

        previous_instances = self._handlers.get(instance.accepted_type, [])
        if instance.accepted_type in self._handlers:
            if len(previous_instances) > 0:
                old_instance = self._get_mimetype_handler_with_name(instance.get_name(),
                                                                    previous_instances)
                if old_instance is not None:
                    replace = options.get('replace', False)
                    if replace is not True:
                        raise DuplicatedMIMETypeHandlerError('There is another registered '
                                                             'mimetype handler with name '
                                                             '[{name}] for accepted type '
                                                             '[{accepted_type}] but "replace" '
                                                             'option is not set, so mimetype '
                                                             'handler [{instance}] could not '
                                                             'be registered.'
                                                             .format(name=instance.get_name(),
                                                                     accepted_type=instance.
                                                                     accepted_type,
                                                                     instance=instance))

                    print_warning('MIMEType handler [{old_instance}] is going '
                                  'to be replaced by [{new_instance}].'
                                  .format(old_instance=old_instance,
                                          new_instance=instance))

                    previous_instances.remove(old_instance)

        previous_instances.append(instance)
        self._set_next_handlers(previous_instances)
        self._handlers[instance.accepted_type] = previous_instances

    def _set_next_handlers(self, mimetype_handlers):
        """
        sets next handler for each mimetype handler in the input list.

        :param list[AbstractMIMETypeHandlerBase] mimetype_handlers: list of mimetype handlers.
        """

        length = len(mimetype_handlers)
        for i in range(length):
            if i == length - 1:
                mimetype_handlers[i].set_next(None)
            else:
                mimetype_handlers[i].set_next(mimetype_handlers[i + 1])

    def _get_mimetype_handler_with_name(self, name, mimetype_handlers):
        """
        gets a mimetype handler with the given name from input mimetype handlers list.

        if not available, it returns None.

        :param str name: mimetype handler name to get its instance.
        :param list[AbstractMIMETypeHandlerBase] mimetype_handlers: list of mimetype handlers.

        :raises DuplicatedMIMETypeHandlerError: duplicated mimetype handler error.

        :rtype: AbstractMIMETypeHandlerBase
        """

        result = [item for item in mimetype_handlers if item.get_name() == name]

        if result is None or len(result) <= 0:
            return None

        if len(result) > 1:
            raise DuplicatedMIMETypeHandlerError('There are multiple mimetype handlers with '
                                                 'name [{name}]. it could be due to a '
                                                 'bug in registering mimetype handlers.'
                                                 .format(name=name))
        if len(result) == 1:
            return result[0]

    def get_mimetype_handlers(self, **options):
        """
        gets all registered mimetype handlers.

        it could filter mimetype handlers for a specific type if provided.
        it only returns the first mimetype handlers for each type, because
        all mimetype handlers for a given type, are chained together.

        :keyword type accepted_type: specifies to get mimetype handlers which are
                                     registered for the accepted type. if not provided,
                                     all mimetype handlers will be returned.

        :rtype: list[AbstractMIMETypeHandlerBase]
        """

        accepted_type = options.get('accepted_type', None)

        if accepted_type is None:
            all_mimetype_handlers = []
            for handler_type in self._handlers:
                all_mimetype_handlers.append(self._handlers[handler_type][0])
            return all_mimetype_handlers
        else:
            mimetype_handlers_keys = [key for key in self._handlers
                                      if issubclass(accepted_type, key)]

            specific_mimetype_handlers = []
            for key in mimetype_handlers_keys:
                specific_mimetype_handlers.append(self._handlers[key][0])

            return specific_mimetype_handlers



# -*- coding: utf-8 -*-
"""
router package.
"""

from pyrin.packaging.base import Package


class RouterPackage(Package):
    """
    router package class.
    """

    NAME = __name__
    DEPENDS = ['pyrin.database.paging']
    COMPONENT_NAME = 'api.router.component'



# -*- coding: utf-8 -*-
"""
template cli module.
"""

import pyrin.template.services as template_services

from pyrin.cli.decorators import cli_invoke, cli_group
from pyrin.core.structs import CLI
from pyrin.template.enumerations import TemplateCLIHandlersEnum


@cli_group('template')
class TemplateCLI(CLI):
    """
    template cli class.

    this class exposes all template cli commands.
    """

    @cli_invoke
    def package(self, **options):
        """
        create a new package for application.

        inputs could be provided within command itself, as well as
        through interactive mode if not provided within command.

        :keyword str package_path: the new package path. it must be a relative
                                   path inside application main package path.

        :keyword str package_class_name: the new package class name.

        :keyword bool help: show the help message for this command.
                            defaults to False if not provided.
        """

        return template_services.create(TemplateCLIHandlersEnum.PACKAGE, **options)

    @cli_invoke
    def empty_package(self, **options):
        """
        create a new empty package for application.

        inputs could be provided within command itself, as well as
        through interactive mode if not provided within command.

        :keyword str package_path: the new package path. it must be a relative
                                   path inside application main package path.

        :keyword str package_class_name: the new package class name.

        :keyword bool help: show the help message for this command.
                            defaults to False if not provided.
        """

        return template_services.create(TemplateCLIHandlersEnum.EMPTY_PACKAGE, **options)



# -*- coding: utf-8 -*-
"""
sentry integrations flask module.
"""

from sentry_sdk.integrations.flask import FlaskIntegration

import pyrin.configuration.services as config_services

from pyrin.logging.sentry.decorators import sentry_integration
from pyrin.logging.sentry.integrations.base import SentryIntegrationBase


@sentry_integration()
class FlaskSentryIntegration(SentryIntegrationBase):
    """
    flask sentry integration class.
    """

    name = 'flask'

    def _configure(self, **options):
        """
        configures this integration and gets the configured object.

        :rtype: FlaskIntegration
        """

        transaction_style = config_services.get('sentry', self.get_name(), 'transaction_style')
        return FlaskIntegration(transaction_style)



# -*- coding: utf-8 -*-
"""
model services module.
"""

from pyrin.application.services import get_component
from pyrin.database.model import ModelPackage


def get_declarative_base():
    """
    gets the application's declarative base class.

    by default it returns `CoreEntity` class.
    if you want to define a new declarative base class in your application
    instead of using `CoreEntity` which is provided by pyrin, you must
    override this method inside your application's `database.model.ModelManager`
    class and return your new declarative base class.
    this is required if you want to enable migrations in your application.
    note that your application must have a unique declarative base type and
    you shouldn't mix the usage of `CoreEntity` and your new declarative base,
    otherwise migrations will not work properly.

    but, if you don't want to use migrations at all, you could just put
    `pyrin.database.migration` into `ignored_packages` list of `packaging.ini`
    file and leave this method unimplemented.

    :rtype: type[BaseEntity]
    """

    return get_component(ModelPackage.COMPONENT_NAME).get_declarative_base()


def get_metadata():
    """
    gets metadata of current declarative base.

    :rtype: MetaData
    """

    return get_component(ModelPackage.COMPONENT_NAME).get_metadata()


def get_tables():
    """
    gets all tables defined in metadata of current declarative base.

    :returns: dict(str name, Table table)
    :rtype: dict
    """

    return get_component(ModelPackage.COMPONENT_NAME).get_tables()


def collect_entities():
    """
    collects all entity classes of application.

    it returns the count of collected entities.

    :rtype: int
    """

    return get_component(ModelPackage.COMPONENT_NAME).collect_entities()


def get_entities():
    """
    gets a tuple of all application collected entities.

    :raises EntitiesAreNotCollectedError: entities are not collected error.

    :rtype: tuple[BaseEntity]
    """

    return get_component(ModelPackage.COMPONENT_NAME).get_entities()


def register_hook(instance):
    """
    registers the given instance into model hooks.

    :param ModelHookBase instance: model hook instance to be registered.

    :raises InvalidModelHookTypeError: invalid model hook type error.
    """

    return get_component(ModelPackage.COMPONENT_NAME).register_hook(instance)


def get_mapper_registry():
    """
    gets the configured mapper registry.

    :rtype: registry
    """

    return get_component(ModelPackage.COMPONENT_NAME).get_mapper_registry()



# -*- coding: utf-8 -*-
"""
response wrappers base module.
"""

from flask import Response
from werkzeug.test import run_wsgi_app

import pyrin.globalization.datetime.services as datetime_services
import pyrin.processor.mimetype.services as mimetype_services

from pyrin.processor.mimetype.enumerations import MIMETypeEnum
from pyrin.processor.response.wrappers.structs import ResponseContext
from pyrin.settings.static import DEFAULT_STATUS_CODE, APPLICATION_ENCODING
from pyrin.processor.response.wrappers.exceptions import InvalidResponseContextKeyNameError, \
    ResponseContextKeyIsAlreadyPresentError, ResponseEnvironRequiredError
from pyrin.processor.exceptions import RequestIDAlreadySetError, \
    RequestDateAlreadySetError, RequestUserAlreadySetError


class CoreResponse(Response):
    """
    core response class.

    this class should be used as base for server response classes.
    """

    # charset of the response.
    charset = APPLICATION_ENCODING

    # class to be used as response context holder.
    response_context_class = ResponseContext

    # default status if none is provided.
    default_status = DEFAULT_STATUS_CODE

    # default mimetype if none is provided.
    default_mimetype = MIMETypeEnum.TEXT

    def __init__(self, response=None, status=None,
                 headers=None, mimetype=None,
                 content_type=None, direct_passthrough=False,
                 **options):
        """
        initializes an instance of CoreResponse.

        :param str | iterable response: a string or response iterable.

        :param str | int status: a string with a status or an integer
                                 with the status code.

        :param list | Headers headers: a list of headers or a
                                       `datastructures.Headers` object.

        :param str mimetype: the mimetype for the response.
                             if not provided, it will be extracted from
                             real response value type. and if not possible,
                             it will be set to `default_mimetype`.

        :param str content_type: the content type for the response.

        :param bool direct_passthrough: if set to True the `iter_encoded` method is not
                                        called before iteration which makes it
                                        possible to pass special iterators through
                                        unchanged.

        :keyword dict original_data: a dict containing the original
                                     data of response before encoding.
                                     this value will be used in logging
                                     to mask critical values.
        """

        if mimetype is None:
            mimetype = self._get_mimetype(response)

        super().__init__(response, status, headers, mimetype,
                         content_type, direct_passthrough)

        self._original_data = None
        self.original_data = options.get('original_data')
        self._request_id = None
        self._request_date = None
        self._response_date = datetime_services.now()
        self._user = None
        self._context = self.response_context_class()

    def __str__(self):
        result = 'request id: "{request_id}", request date: "{request_date}", ' \
                 'response date: "{response_date}", user: "{user}", ' \
                 'status_code: "{status_code}"'
        return result.format(response_date=self._response_date,
                             request_id=self.request_id,
                             request_date=self.request_date,
                             user=self.user,
                             status_code=self.status_code)

    def __hash__(self):
        return hash(self.request_id)

    def _get_mimetype(self, response):
        """
        gets the correct mimetype of given response object.

        it gets the `default_mimetype` if it fails to detect.

        :param str | iterable response: a string or response iterable.

        :returns: mimetype name
        :rtype: str
        """

        mimetype = mimetype_services.get_mimetype(response)
        if mimetype is None:
            mimetype = self.default_mimetype

        return mimetype

    def add_context(self, key, value, **options):
        """
        adds the given key/value pair into current response context.

        :param str key: key name to be added.
        :param object value: value to be added.

        :keyword bool replace: specifies that if a key with the same name
                               is already present, replace it. otherwise
                               raise an error. defaults to False if not provided.

        :raises InvalidResponseContextKeyNameError: invalid response context key name error.
        :raises ResponseContextKeyIsAlreadyPresentError: response context key is
                                                         already present error.
        """

        if key in (None, '') or key.isspace():
            raise InvalidResponseContextKeyNameError('Response context key must be provided.')

        if key in self._context:
            replace = options.get('replace', None)
            if replace is None:
                replace = False

            if replace is not True:
                raise ResponseContextKeyIsAlreadyPresentError('A response context with key '
                                                              '[{key}] is already present '
                                                              'and "replace" option is not set.'
                                                              .format(key=key))
        self._context[key] = value

    def get_context(self, key, default=None):
        """
        gets the value for given key from current response context.

        it gets the default value if key is not present in the response context.

        :param str key: key name to get its value.
        :param object default: a value to be returned if the provided
                               key is not present in response context.

        :rtype: object
        """

        return self._context.get(key, default)

    def remove_context(self, key):
        """
        removes the specified key from current response context if available.

        :param str key: key name to be removed from response context.
        """

        self._context.pop(key, None)

    @classmethod
    def force_type(cls, response, environ=None):
        """
        enforces that the wsgi response is a response object of the current type.

        this method can enforce a given response type, and it will also
        convert arbitrary wsgi callables into response objects if an environ
        is provided.

        this is especially useful if you want to post-process responses in
        the main dispatcher and use functionality provided by your subclass.

        keep in mind that this will modify response objects in place if
        possible.

        :param object response: a response object or wsgi application.
        :param dict environ: a wsgi environment object.

        :raises ResponseEnvironRequiredError: response environ required error.

        :returns: a response object.
        :rtype: CoreResponse
        """

        if not isinstance(response, CoreResponse):
            if environ is None:
                raise ResponseEnvironRequiredError('Cannot convert WSGI application into '
                                                   'response object without an environ.')

            response = CoreResponse(*run_wsgi_app(response, environ))

        response.__class__ = cls
        return response

    @property
    def safe_content_length(self):
        """
        gets current response's content length if available, otherwise returns 0.

        :rtype: int
        """

        return self.content_length or 0

    @property
    def request_id(self):
        """
        gets current response's request id.

        :rtype: uuid.UUID
        """

        return self._request_id

    @request_id.setter
    def request_id(self, request_id):
        """
        sets current response's request id.

        :param uuid.UUID request_id: request id to be set.

        :raises RequestIDAlreadySetError: request id already set error.
        """

        if self._request_id is not None:
            raise RequestIDAlreadySetError('Request id for current response '
                                           'has been already set.')

        self._request_id = request_id

    @property
    def request_date(self):
        """
        gets current response's request date.

        :rtype: datetime
        """

        return self._request_date

    @request_date.setter
    def request_date(self, request_date):
        """
        sets current response's request date.

        :param datetime request_date: request date to be set.

        :raises RequestDateAlreadySetError: request date already set error.
        """

        if self._request_date is not None:
            raise RequestDateAlreadySetError('Request date for current response '
                                             'has been already set.')

        self._request_date = request_date

    @property
    def user(self):
        """
        gets current response's user.

        :rtype: object
        """

        return self._user

    @user.setter
    def user(self, user):
        """
        sets current response's user.

        :param object user: user to be set.

        :raises RequestUserAlreadySetError: request user already set error.
        """

        if self._user is not None:
            raise RequestUserAlreadySetError('Request user for current response '
                                             'has been already set.')

        self._user = user

    @property
    def original_data(self):
        """
        gets the original data of this response.

        it returns a dict if the original data was a dict, otherwise
        it gets the response data as text.

        :rtype: dict | str
        """

        return self._original_data or self.get_data(as_text=True)

    @original_data.setter
    def original_data(self, data):
        """
        sets the response original data if it is of dict type.

        otherwise sets None for this value.
        this value will be used for response logging to let loggers mask critical values.

        :param dict data: data to be registered.
        """

        if isinstance(data, dict):
            self._original_data = dict(data)



# -*- coding: utf-8 -*-
"""
datetime services module.
"""

from pyrin.application.services import get_component
from pyrin.globalization.datetime import DateTimePackage


def now(server=True, timezone=None):
    """
    gets the current datetime based on requested timezone.

    :param bool server: if set to True, server timezone will be used.
                        if set to False, client timezone will be used.
                        defaults to True.

    :param str timezone: timezone name to get datetime based on it.
                         if provided, the value of `server` input
                         will be ignored. defaults to None.

    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).now(server, timezone)


def client_now():
    """
    gets the current datetime based on client timezone.

    this is a helper method to let get the client datetime
    without providing value to `now` method.

    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).client_now()


def today(server=True, timezone=None):
    """
    gets the current date based on requested timezone.

    :param bool server: if set to True, server timezone will be used.
                        if set to False, client timezone will be used.
                        defaults to True.

    :param str timezone: timezone name to get date based on it.
                         if provided, the value of `server` input
                         will be ignored. defaults to None.

    :rtype: date
    """

    return get_component(DateTimePackage.COMPONENT_NAME).today(server, timezone)


def client_today():
    """
    gets the current date based on client timezone.

    this is a helper method to let get the client date
    without providing value to `today` method.

    :rtype: date
    """

    return get_component(DateTimePackage.COMPONENT_NAME).client_today()


def current_year(server=True, timezone=None):
    """
    gets the current year based on requested timezone.

    :param bool server: if set to True, server timezone will be used.
                        if set to False, client timezone will be used.
                        defaults to True.

    :param str timezone: timezone name to get year based on it.
                         if provided, the value of `server` input
                         will be ignored. defaults to None.

    :rtype: int
    """

    return get_component(DateTimePackage.COMPONENT_NAME).current_year(server, timezone)


def current_client_year():
    """
    gets the current year based on client timezone.

    this is a helper method to let get the client year
    without providing value to `current_year` method.

    :rtype: int
    """

    return get_component(DateTimePackage.COMPONENT_NAME).current_client_year()


def get_default_client_timezone():
    """
    gets the default client timezone.

    :rtype: tzinfo
    """

    return get_component(DateTimePackage.COMPONENT_NAME).get_default_client_timezone()


def get_current_timezone(server):
    """
    gets the current timezone for server or client.

    :param bool server: if set to True, server timezone will be returned.
                        if set to False, client timezone will be returned.

    :rtype: tzinfo
    """

    return get_component(DateTimePackage.COMPONENT_NAME).get_current_timezone(server)


def get_timezone(timezone):
    """
    gets the timezone based on given timezone name.

    :param str timezone: timezone name.

    :rtype: tzinfo
    """

    return get_component(DateTimePackage.COMPONENT_NAME).get_timezone(timezone)


def get_timezone_name(server):
    """
    gets the server or client timezone name.

    :param bool server: if set to True, server timezone name will be returned.
                        if set to False, client timezone name will be returned.

    :rtype: str
    """

    return get_component(DateTimePackage.COMPONENT_NAME).get_timezone_name(server)


def convert(value, to_server, from_server=None):
    """
    converts the given datetime between server and client timezones.

    :param datetime value: value to be converted.

    :param bool to_server: specifies that value must be normalized
                           to server timezone. if set to False, it
                           will be normalized to client timezone.

    :param bool from_server: specifies that value must be normalized
                             from server timezone. if set to False, it
                             will be normalized from client timezone.
                             if not provided, it will be set to opposite
                             of `to_server` value.
    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).convert(value, to_server,
                                                                 from_server=from_server)


def convert_to_utc(value, from_server):
    """
    converts the given datetime to utc.

    :param datetime value: value to be converted.

    :param bool from_server: specifies that value must be normalized
                             from server timezone. if set to False, it
                             will be normalized from client timezone.

    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).convert_to_utc(value, from_server)


def convert_from_utc(value, to_server):
    """
    converts the given datetime from utc.

    :param datetime value: value to be converted.

    :param bool to_server: specifies that value must be normalized
                           to server timezone. if set to False, it
                           will be normalized to client timezone.

    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).convert_from_utc(value, to_server)


def as_timezone(value, server):
    """
    gets the result of `astimezone` on the given value.

    :param datetime value: value to get normalized.

    :param bool server: if set to True, server timezone name will be used.
                        if set to False, client timezone name will be used.

    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).as_timezone(value, server)


def normalize(value, server):
    """
    normalizes input value using server or client current timezone and returns it.

    :param datetime value: value to get normalized.
    :param bool server: specifies that server or client timezone must used.

    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).normalize(value, server)


def localize(value, server):
    """
    localizes input datetime with current timezone.

    input value should not have a timezone info.

    :param datetime value: value to be localized.
    :param bool server: specifies that server or client timezone must used.

    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).localize(value, server)


def try_add_timezone(value, server):
    """
    adds the current timezone info into input value if it has no timezone info.

    :param datetime value: value to add timezone info into it.
    :param bool server: specifies that server or client timezone must be added.

    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).try_add_timezone(value, server)


def to_datetime_string(value, to_server, from_server=None):
    """
    gets the datetime string representation of input value.

    if the value has no timezone info, it adds the client or server
    timezone info based on `from_server` value.

    example: `2015-12-24T22:40:15+01:00`

    :param datetime value: input object to be converted.

    :param bool to_server: specifies that value must be normalized
                           to server timezone. if set to False, it
                           will be normalized to client timezone.

    :param bool from_server: specifies that value must be normalized
                             from server timezone. if set to False, it
                             will be normalized from client timezone.
                             if not provided, it will be set to opposite
                             of `to_server` value.

    :rtype: str
    """

    return get_component(DateTimePackage.COMPONENT_NAME).to_datetime_string(
        value, to_server, from_server=from_server)


def to_date_string(value):
    """
    gets the date string representation of input value.

    example: `2015-12-24`

    :param datetime | date value: input object to be converted.

    :rtype: str
    """

    return get_component(DateTimePackage.COMPONENT_NAME).to_date_string(value)


def to_time_string(value, to_server, from_server=None):
    """
    gets the time string representation of input value.

    if the value is a datetime and has no timezone info, it adds
    the client or server timezone info based on `from_server` value.

    example: `23:40:15`

    :param datetime | time value: input object to be converted.

    :param bool to_server: specifies that value must be normalized
                           to server timezone. if set to False, it
                           will be normalized to client timezone.

    :param bool from_server: specifies that value must be normalized
                             from server timezone. if set to False, it
                             will be normalized from client timezone.
                             if not provided, it will be set to opposite
                             of `to_server` value.

    :rtype: str
    """

    return get_component(DateTimePackage.COMPONENT_NAME).to_time_string(value, to_server,
                                                                        from_server=from_server)


def to_datetime(value, to_server, from_server=None):
    """
    converts the input value to it's equivalent python datetime.

    if the value has no timezone info, it adds the client or server
    timezone info based on `from_server` value.

    :param str value: string representation of datetime to be converted.

    :param bool to_server: specifies that value must be normalized
                           to server timezone. if set to False, it
                           will be normalized to client timezone.

    :param bool from_server: specifies that value must be normalized
                             from server timezone. if set to False, it
                             will be normalized from client timezone.
                             if not provided, it will be set to opposite
                             of `to_server` value.

    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).to_datetime(value, to_server,
                                                                     from_server=from_server)


def to_date(value):
    """
    converts the input value to it's equivalent python date.

    :param str value: string representation of date to be converted.

    :rtype: date
    """

    return get_component(DateTimePackage.COMPONENT_NAME).to_date(value)


def to_time(value):
    """
    converts the input value to it's equivalent python time.

    :param str value: string representation of time to be converted.

    :rtype: time
    """

    return get_component(DateTimePackage.COMPONENT_NAME).to_time(value)


def timezone_exists(timezone_name):
    """
    gets a value indicating that a timezone with the given name exists.

    :param str timezone_name: timezone name to check for existence.

    :rtype: bool
    """

    return get_component(DateTimePackage.COMPONENT_NAME).timezone_exists(timezone_name)


def get_timestamp(value, date_sep='-', main_sep=' ',
                  time_sep=':', microsecond=False):
    """
    gets the timestamp with specified separators for given datetime.

    default format is `YYYY-MM-DD HH:mm:SS`.

    :param datetime value: datetime value to get its timestamp.

    :param str date_sep: a separator to put between date elements.
                         if set to None, no separator will be used.

    :param str main_sep: a separator to put between date and time part.
                         if set to None, no separator will be used.

    :param str time_sep: a separator to put between time elements.
                         if set to None, no separator will be used.

    :param bool microsecond: specifies that timestamp must include microseconds.
                             defaults to False if not provided.

    :rtype: str
    """

    return get_component(DateTimePackage.COMPONENT_NAME).get_timestamp(value, date_sep,
                                                                       main_sep, time_sep,
                                                                       microsecond)


def get_current_timestamp(date_sep='-', main_sep=' ',
                          time_sep=':', server=True,
                          timezone=None, microsecond=False):
    """
    gets the current timestamp with specified separators based on requested timezone.

    default format is `YYYY-MM-DD HH:mm:SS`.

    :param str date_sep: a separator to put between date elements.
                         if set to None, no separator will be used.

    :param str main_sep: a separator to put between date and time part.
                         if set to None, no separator will be used.

    :param str time_sep: a separator to put between time elements.
                         if set to None, no separator will be used.

    :param bool server: if set to True, server timezone will be used.
                        if set to False, client timezone will be used.
                        defaults to True.

    :param str timezone: timezone name to get datetime based on it.
                         if provided, the value of `server` input
                         will be ignored. defaults to None.

    :param bool microsecond: specifies that timestamp must include microseconds.
                             defaults to False if not provided.

    :rtype: str
    """

    return get_component(DateTimePackage.COMPONENT_NAME).get_current_timestamp(date_sep,
                                                                               main_sep,
                                                                               time_sep,
                                                                               server,
                                                                               timezone,
                                                                               microsecond)


def datetime(year, month, day, hour=0, minute=0,
             second=0, microsecond=0, fold=0,
             server=True, timezone=None):
    """
    gets a new datetime with given inputs and requested timezone.

    :param int year: year.
    :param int month: month.
    :param int day: day.
    :param int hour: hour.
    :param int minute: minute.
    :param int second: second.
    :param int microsecond: microsecond.

    :param int fold: used to disambiguate wall times during a repeated
                     interval. it could be set to 0 or 1.

    :param bool server: if set to True, server timezone will be used.
                        if set to False, client timezone will be used.
                        defaults to True.

    :param str timezone: timezone name to be used.
                         if provided, the value of `server` input
                         will be ignored. defaults to None.
    :rtype: datetime
    """

    return get_component(DateTimePackage.COMPONENT_NAME).datetime(year, month, day,
                                                                  hour, minute, second,
                                                                  microsecond, fold,
                                                                  server, timezone)



# -*- coding: utf-8 -*-
"""
processor exceptions module.
"""

from pyrin.core.exceptions import CoreException


class ProcessorException(CoreException):
    """
    processor exception.
    """
    pass


class RequestIDAlreadySetError(ProcessorException):
    """
    request id already set error.
    """
    pass


class RequestDateAlreadySetError(ProcessorException):
    """
    request date already set error.
    """
    pass


class RequestUserAlreadySetError(ProcessorException):
    """
    request user already set error.
    """
    pass



# -*- coding: utf-8 -*-
"""
utils environment module.
"""

import os
import sys
import platform


def get_os_name():
    """
    gets current os name.

    :rtype: str
    """

    return platform.uname()[0].lower()


def is_windows():
    """
    gets a value indicating that underlying operating system is windows based.

    :rtype: bool
    """

    return 'win' in get_os_name()


def set_python_path(python_path):
    """
    adds the given path into `PYTHONPATH` variable.

    :param str python_path: python path to set.
    """

    python_path = os.path.abspath(python_path)
    os.environ['PYTHONPATH'] = python_path
    sys.path.append(python_path)



# -*- coding: utf-8 -*-
"""
router manager module.
"""

import pyrin.application.services as application_services
import pyrin.utils.misc as misc_utils

from pyrin.core.structs import Manager
from pyrin.api.router import RouterPackage
from pyrin.api.router.handlers.base import RouteBase
from pyrin.api.router.handlers.public import PublicRoute, PublicTemporaryRoute
from pyrin.api.router.handlers.protected import ProtectedRoute, FreshProtectedRoute, \
    ProtectedTemporaryRoute, FreshProtectedTemporaryRoute
from pyrin.api.router.exceptions import RouteAuthenticationMismatchError, \
    InvalidCustomRouteTypeError


class RouterManager(Manager):
    """
    router manager class.
    """

    package_class = RouterPackage

    def create_route(self, rule, **options):
        """
        creates the appropriate route based on the input parameters.

        this method acts as a route factory and its corresponding service
        will be registered as default route factory on application startup.

        :param str rule: unique url rule to register this route for.
                         routes with duplicated urls and http methods will be
                         overwritten if `replace=True` option is provided.
                         otherwise an error will be raised.

        :keyword bool authenticated: specifies that this route could not be accessed
                                     if the requester has not been authenticated.
                                     defaults to True if not provided.

        :keyword bool fresh_auth: specifies that this route could not be accessed
                                  if the requester has not a fresh authentication.
                                  fresh authentication means an authentication that
                                  has been done by providing user credentials to server.
                                  defaults to False if not provided.

        :keyword PermissionBase | tuple[PermissionBase] permissions: all required permissions
                                                                     to access this route.

        :keyword str | tuple[str] methods: http methods that this route could handle.
                                           if not provided, defaults to `GET`, `HEAD`
                                           and `OPTIONS`.

        :keyword function view_function: a function to be called on accessing this route.

        :keyword str endpoint: the endpoint for the registered url rule.

        :keyword dict defaults: an optional dict with defaults for other rules with the
                                same endpoint. this is a bit tricky but useful if you
                                want to have unique urls.

        :keyword str subdomain: the subdomain rule string for this rule. If not specified the
                                rule only matches for the `default_subdomain` of the map. if
                                the map is not bound to a subdomain this feature is disabled.

        :keyword bool strict_slashes: override the `Map` setting for `strict_slashes` only for
                                      this rule. if not specified the `Map` setting is used.

        :keyword bool merge_slashes: override `Map.merge_slashes` for this rule.

        :keyword bool build_only: set this to True and the rule will never match but will
                                  create a url that can be build. this is useful if you have
                                  resources on a subdomain or folder that are not handled by
                                  the WSGI application (like static data)

        :keyword str | callable redirect_to: if given this must be either a string
                                             or callable. in case of a callable it's
                                             called with the url adapter that
                                             triggered the match and the values
                                             of the url as keyword arguments and has
                                             to return the target for the redirect,
                                             otherwise it has to be a string with
                                             placeholders in rule syntax.

        :keyword bool alias: if enabled this rule serves as an alias for another rule with
                             the same endpoint and arguments.

        :keyword str host: if provided and the url map has host matching enabled this can be
                           used to provide a match rule for the whole host. this also means
                           that the subdomain feature is disabled.

        :keyword bool websocket: if set to True, this rule is only matches for
                                 websocket (`ws://`, `wss://`) requests. by default,
                                 rules will only match for http requests.
                                 defaults to False if not provided.

        :keyword int max_content_length: max content length that this route could handle,
                                         in bytes. if not provided, it will be set to
                                         `restricted_max_content_length` api config key.
                                         note that this value should be lesser than or equal
                                         to `max_content_length` api config key, otherwise
                                         it will cause an error.

        :keyword int status_code: status code to be returned on successful responses.
                                  defaults to corresponding status code of request's
                                  http method if not provided.

        :note status_code: it could be a value from `InformationResponseCodeEnum`
                           or `SuccessfulResponseCodeEnum` or `RedirectionResponseCodeEnum`.

        :keyword bool strict_status: specifies that it should only consider
                                     the status code as processed if it is from
                                     `InformationResponseCodeEnum` or
                                     `SuccessfulResponseCodeEnum` or
                                     `RedirectionResponseCodeEnum` values. otherwise
                                     all codes from `INFORMATION_CODE_MIN`
                                     to `INFORMATION_CODE_MAX` or from
                                     `SUCCESS_CODE_MIN` to `SUCCESS_CODE_MAX`
                                     or from `REDIRECTION_CODE_MIN` to
                                     `REDIRECTION_CODE_MAX` will be considered
                                     as processed. defaults to True if not provided.

        :keyword ResultSchema | type[ResultSchema] result_schema: result schema to be used
                                                                  to filter results. it could
                                                                  be an instance or a type
                                                                  of `ResultSchema` class.

        :keyword bool indexed: specifies that list results must
                               include an extra field as row index.
                               the name of the index field and the initial value
                               of index could be provided by `index_name` and
                               `start_index` respectively. `indexed` keyword has
                               only effect if the returning result contains a list
                               of objects.

        :keyword str index_name: name of the extra field to contain
                                 the row index of each result. if not provided
                                 defaults to `row_num` value.

        :keyword int start_index: the initial value of row index. if not
                                  provided, starts from 1.

        :keyword SECURE_TRUE | SECURE_FALSE readable: specifies that any column or attribute
                                                      which has `allow_read=False` or its name
                                                      starts with underscore `_`, should not
                                                      be included in result dict. defaults
                                                      to `SECURE_TRUE` if not provided. it
                                                      will be used only for entity conversion.
                                                      this value will override the
                                                      corresponding value of `result_schema`
                                                      if provided.

        :keyword int depth: a value indicating the depth for conversion.
                            for example if entity A has a relationship with
                            entity B and there is a list of B in A, if `depth=0`
                            is provided, then just columns of A will be available
                            in result dict, but if `depth=1` is provided, then all
                            B entities in A will also be included in the result dict.
                            actually, `depth` specifies that relationships in an
                            entity should be followed by how much depth.
                            note that, if `columns` is also provided, it is required to
                            specify relationship property names in provided columns.
                            otherwise they won't be included even if `depth` is provided.
                            defaults to `default_depth` value of database config store.
                            please be careful on increasing `depth`, it could fail
                            application if set to higher values. choose it wisely.
                            normally the maximum acceptable `depth` would be 2 or 3.
                            there is a hard limit for max valid `depth` which is set
                            in `ConverterMixin.MAX_DEPTH` class variable. providing higher
                            `depth` value than this limit, will cause an error.
                            it will be used only for entity conversion.
                            this value will override the corresponding value of
                            `result_schema` if provided.

        :keyword bool no_cache: a value indicating that the response returning from this route
                                must have a `Cache-Control: no-cache` header. this header will
                                be automatically added. defaults to False if not provided.

        :keyword int request_limit: number of allowed requests to this
                                    route before it unregisters itself.
                                    defaults to None if not Provided.

        :keyword int lifetime: number of seconds in which this route must remain
                               responsive after initial registration. after this
                               period, the route will unregister itself.
                               defaults to None if not provided.

        :note request_limit, lifetime: if both of these values are provided, this
                                       route will unregister itself if any of
                                       these two conditions are met.

        :keyword bool paged: specifies that this route should return paginated results.
                             defaults to False if not provided.

        :keyword int page_size: default page size for this route.
                                defaults to `default_page_size` from
                                `database` config store if not provided.

        :keyword int max_page_size: maximum page size that client is allowed
                                    to request for this route. defaults to
                                    `max_page_size` from `database` configs store
                                    if not provided.

        :keyword bool cors_enabled: specifies that cross origin resource sharing is enabled.
                                    if not provided, it will be get from cors config store.

        :keyword bool cors_always_send: specifies that cors headers must be included in
                                        response even if the request does not have origin header.
                                        if not provided, it will be get from cors config store.

        :keyword list[str] cors_allowed_origins: a list of extra allowed origins to be used
                                                 in conjunction with default allowed ones.

        :keyword list[str] cors_exposed_headers: extra exposed headers to be combined
                                                 with default ones.

        :keyword list[str] cors_allowed_headers: extra allowed headers to be combined
                                                 with default ones.

        :keyword bool cors_allow_credentials: specifies that browsers are allowed to pass
                                              response headers to front-end javascript code
                                              if the route is authenticated.
                                              if not provided, it will be get from cors config
                                              store.

        :keyword int cors_max_age: maximum number of seconds to cache results.
                                   if not provided, it will be get from cors config store.

        :raises InvalidCustomRouteTypeError: invalid custom route type error.
        :raises RouteAuthenticationMismatchError: route authentication mismatch error.
        :raises PageSizeLimitError: page size limit error.
        :raises MaxContentLengthLimitMismatchError: max content length limit mismatch error.
        :raises InvalidViewFunctionTypeError: invalid view function type error.
        :raises InvalidResultSchemaTypeError: invalid result schema type error.
        :raises InvalidResponseStatusCodeError: invalid response status code error.

        :rtype: RouteBase
        """

        route = self._create_route(rule, **options)
        if route is not None:
            if not isinstance(route, RouteBase):
                route_name = misc_utils.try_get_fully_qualified_name(route)
                base_name = misc_utils.try_get_fully_qualified_name(RouteBase)
                raise InvalidCustomRouteTypeError('Custom route [{route}] is not '
                                                  'an instance of [{base}].'
                                                  .format(route=route_name, base=base_name))
            return route

        authenticated = options.get('authenticated')
        if authenticated is None:
            authenticated = True

        fresh_auth = options.get('fresh_auth')
        if fresh_auth is None:
            fresh_auth = False

        request_limit = options.get('request_limit')
        lifetime = options.get('lifetime')
        temporary = request_limit is not None or lifetime is not None

        if authenticated is False and fresh_auth is False and temporary is False:
            return PublicRoute(rule, **options)
        elif authenticated is False and fresh_auth is False and temporary is True:
            return PublicTemporaryRoute(rule, **options)
        elif authenticated is True and fresh_auth is False and temporary is False:
            return ProtectedRoute(rule, **options)
        elif authenticated is True and fresh_auth is False and temporary is True:
            return ProtectedTemporaryRoute(rule, **options)
        elif authenticated is True and fresh_auth is True and temporary is False:
            return FreshProtectedRoute(rule, **options)
        elif authenticated is True and fresh_auth is True and temporary is True:
            return FreshProtectedTemporaryRoute(rule, **options)
        else:
            raise RouteAuthenticationMismatchError('[authenticated={auth}] and '
                                                   '[fresh_auth={fresh}] and '
                                                   '[temporary={temp}] in route '
                                                   '[{route}] are incompatible.'
                                                   .format(auth=authenticated,
                                                           fresh=fresh_auth,
                                                           temp=temporary,
                                                           route=rule))

    def _create_route(self, rule, **options):
        """
        creates the appropriate route based on the input parameters.

        this method is intended to be overridden by subclasses to provided
        custom `RouteBase` types. it should always return a descendant object
         of `RouteBase` or `None`.

        :param str rule: unique url rule to register this route for.
                         routes with duplicated urls and http methods will be
                         overwritten if `replace=True` option is provided.
                         otherwise an error will be raised.

        :keyword bool authenticated: specifies that this route could not be accessed
                                     if the requester has not been authenticated.
                                     defaults to True if not provided.

        :keyword bool fresh_auth: specifies that this route could not be accessed
                                  if the requester has not a fresh authentication.
                                  fresh authentication means an authentication that
                                  has been done by providing user credentials to server.
                                  defaults to False if not provided.

        :keyword PermissionBase | tuple[PermissionBase] permissions: all required permissions
                                                                     to access this route.

        :keyword str | tuple[str] methods: http methods that this route could handle.
                                           if not provided, defaults to `GET`, `HEAD`
                                           and `OPTIONS`.

        :keyword function view_function: a function to be called on accessing this route.

        :keyword str endpoint: the endpoint for the registered url rule.

        :keyword dict defaults: an optional dict with defaults for other rules with the
                                same endpoint. this is a bit tricky but useful if you
                                want to have unique urls.

        :keyword str subdomain: the subdomain rule string for this rule. If not specified the
                                rule only matches for the `default_subdomain` of the map. if
                                the map is not bound to a subdomain this feature is disabled.

        :keyword bool strict_slashes: override the `Map` setting for `strict_slashes` only for
                                      this rule. if not specified the `Map` setting is used.

        :keyword bool merge_slashes: override `Map.merge_slashes` for this rule.

        :keyword bool build_only: set this to True and the rule will never match but will
                                  create a url that can be build. this is useful if you have
                                  resources on a subdomain or folder that are not handled by
                                  the WSGI application (like static data)

        :keyword str | callable redirect_to: if given this must be either a string
                                             or callable. in case of a callable it's
                                             called with the url adapter that
                                             triggered the match and the values
                                             of the url as keyword arguments and has
                                             to return the target for the redirect,
                                             otherwise it has to be a string with
                                             placeholders in rule syntax.

        :keyword bool alias: if enabled this rule serves as an alias for another rule with
                             the same endpoint and arguments.

        :keyword str host: if provided and the url map has host matching enabled this can be
                           used to provide a match rule for the whole host. this also means
                           that the subdomain feature is disabled.

        :keyword bool websocket: if set to True, this rule is only matches for
                                 websocket (`ws://`, `wss://`) requests. by default,
                                 rules will only match for http requests.
                                 defaults to False if not provided.

        :keyword int max_content_length: max content length that this route could handle,
                                         in bytes. if not provided, it will be set to
                                         `restricted_max_content_length` api config key.
                                         note that this value should be lesser than or equal
                                         to `max_content_length` api config key, otherwise
                                         it will cause an error.

        :keyword int status_code: status code to be returned on successful responses.
                                  defaults to corresponding status code of request's
                                  http method if not provided.

        :note status_code: it could be a value from `InformationResponseCodeEnum`
                           or `SuccessfulResponseCodeEnum` or `RedirectionResponseCodeEnum`.

        :keyword bool strict_status: specifies that it should only consider
                                     the status code as processed if it is from
                                     `InformationResponseCodeEnum` or
                                     `SuccessfulResponseCodeEnum` or
                                     `RedirectionResponseCodeEnum` values. otherwise
                                     all codes from `INFORMATION_CODE_MIN`
                                     to `INFORMATION_CODE_MAX` or from
                                     `SUCCESS_CODE_MIN` to `SUCCESS_CODE_MAX`
                                     or from `REDIRECTION_CODE_MIN` to
                                     `REDIRECTION_CODE_MAX` will be considered
                                     as processed. defaults to True if not provided.

        :keyword ResultSchema | type[ResultSchema] result_schema: result schema to be used
                                                                  to filter results. it could
                                                                  be an instance or a type
                                                                  of `ResultSchema` class.

        :keyword bool indexed: specifies that list results must
                               include an extra field as row index.
                               the name of the index field and the initial value
                               of index could be provided by `index_name` and
                               `start_index` respectively. `indexed` keyword has
                               only effect if the returning result contains a list
                               of objects.

        :keyword str index_name: name of the extra field to contain
                                 the row index of each result. if not provided
                                 defaults to `row_num` value.

        :keyword int start_index: the initial value of row index. if not
                                  provided, starts from 1.

        :keyword SECURE_TRUE | SECURE_FALSE readable: specifies that any column or attribute
                                                      which has `allow_read=False` or its name
                                                      starts with underscore `_`, should not
                                                      be included in result dict. defaults to
                                                      `SECURE_TRUE` if not provided. it will
                                                      be used only for entity conversion.
                                                      this value will override the
                                                      corresponding value of `result_schema`
                                                      if provided.

        :keyword int depth: a value indicating the depth for conversion.
                            for example if entity A has a relationship with
                            entity B and there is a list of B in A, if `depth=0`
                            is provided, then just columns of A will be available
                            in result dict, but if `depth=1` is provided, then all
                            B entities in A will also be included in the result dict.
                            actually, `depth` specifies that relationships in an
                            entity should be followed by how much depth.
                            note that, if `columns` is also provided, it is required to
                            specify relationship property names in provided columns.
                            otherwise they won't be included even if `depth` is provided.
                            defaults to `default_depth` value of database config store.
                            please be careful on increasing `depth`, it could fail
                            application if set to higher values. choose it wisely.
                            normally the maximum acceptable `depth` would be 2 or 3.
                            there is a hard limit for max valid `depth` which is set
                            in `ConverterMixin.MAX_DEPTH` class variable. providing higher
                            `depth` value than this limit, will cause an error.
                            it will be used only for entity conversion.
                            this value will override the corresponding value of
                            `result_schema` if provided.

        :keyword bool no_cache: a value indicating that the response returning from this route
                                must have a `Cache-Control: no-cache` header. this header will
                                be automatically added. defaults to False if not provided.

        :keyword int request_limit: number of allowed requests to this
                                    route before it unregisters itself.
                                    defaults to None if not Provided.

        :keyword int lifetime: number of seconds in which this route must remain
                               responsive after initial registration. after this
                               period, the route will unregister itself.
                               defaults to None if not provided.

        :note request_limit, lifetime: if both of these values are provided, this
                                       route will unregister itself if any of
                                       these two conditions are met.

        :keyword bool paged: specifies that this route should return paginated results.
                             defaults to False if not provided.

        :keyword int page_size: default page size for this route.
                                defaults to `default_page_size` from
                                `database` config store if not provided.

        :keyword int max_page_size: maximum page size that client is allowed
                                    to request for this route. defaults to
                                    `max_page_size` from `database` configs store
                                    if not provided.

        :keyword bool cors_enabled: specifies that cross origin resource sharing is enabled.
                                    if not provided, it will be get from cors config store.

        :keyword bool cors_always_send: specifies that cors headers must be included in
                                        response even if the request does not have origin header.
                                        if not provided, it will be get from cors config store.

        :keyword list[str] cors_allowed_origins: a list of extra allowed origins to be used
                                                 in conjunction with default allowed ones.

        :keyword list[str] cors_exposed_headers: extra exposed headers to be combined
                                                 with default ones.

        :keyword list[str] cors_allowed_headers: extra allowed headers to be combined
                                                 with default ones.

        :keyword bool cors_allow_credentials: specifies that browsers are allowed to pass
                                              response headers to front-end javascript code
                                              if the route is authenticated.
                                              if not provided, it will be get from cors config
                                              store.

        :keyword int cors_max_age: maximum number of seconds to cache results.
                                   if not provided, it will be get from cors config store.

        :raises PageSizeLimitError: page size limit error.
        :raises MaxContentLengthLimitMismatchError: max content length limit mismatch error.
        :raises InvalidViewFunctionTypeError: invalid view function type error.
        :raises InvalidResultSchemaTypeError: invalid result schema type error.
        :raises InvalidResponseStatusCodeError: invalid response status code error.

        :rtype: RouteBase
        """

        return None

    def add_route(self, url, view_func=None,
                  provide_automatic_options=None, **options):
        """
        connects a url rule. the provided view_func will be registered with the endpoint.

        if there is another rule with the same url and http methods and `replace=True`
        option is provided, it will be replaced. otherwise an error will be raised.

        a note about endpoint. pyrin will handle endpoint generation on its own.
        so there is no endpoint parameter in this method's signature.
        this is required to be able to handle uniqueness of endpoints and managing them.
        despite flask, pyrin will not require you to define view functions with unique names.
        you could define view functions with the same name in different modules. but to
        ensure the uniqueness of endpoints, pyrin will use the fully qualified name
        of function as endpoint. for example: `pyrin.api.services.create_route`.
        so you could figure out endpoint for any view function to use it in places
        like `url_for` method.

        pyrin handles endpoints this way to achieve these two important features:

        1. dismissal of the need for view function name uniqueness.
           when application grows, it's nonsense to be forced to have
           unique view function names at application level.

        2. managing routes with the same url and http methods, and informing
           the developer about them to prevent accidental replacements. and also
           providing a way to replace a route by another route with the same url
           and http methods if that is what developer actually wants.
           when application grows, it becomes a point of failure when you have no
           idea that you've registered a similar route in multiple places and only
           one of them will be get called based on registration order.

        :param str url: the url rule as string.

        :param function view_func: the function to call when serving a request to the
                                   provided endpoint.

        :param bool provide_automatic_options: controls whether the `OPTIONS` method should be
                                               added automatically.
                                               this can also be controlled by setting the
                                               `view_func.provide_automatic_options = False`
                                               before adding the rule.

        :keyword str | tuple[str] methods: http methods that this rule should handle.
                                           if not provided, defaults to `GET`.

        :keyword PermissionBase | tuple[PermissionBase] permissions: all required permissions
                                                                     for accessing this route.

        :keyword bool authenticated: specifies that this route could not be accessed
                                     if the requester has not been authenticated.
                                     defaults to True if not provided.

        :keyword bool fresh_auth: specifies that this route could not be accessed
                                  if the requester has not a fresh authentication.
                                  fresh authentication means an authentication that
                                  has been done by providing user credentials to
                                  server. defaults to False if not provided.

        :keyword bool replace: specifies that this route must replace any existing
                               route with the same url and http methods or raise
                               an error if not provided. defaults to False.

        :keyword dict defaults: an optional dict with defaults for other rules with the
                                same endpoint. this is a bit tricky but useful if you
                                want to have unique urls.

        :keyword str subdomain: the subdomain rule string for this rule. If not specified the
                                rule only matches for the `default_subdomain` of the map. if
                                the map is not bound to a subdomain this feature is disabled.

        :keyword bool strict_slashes: override the `Map` setting for `strict_slashes` only for
                                      this rule. if not specified the `Map` setting is used.

        :keyword bool merge_slashes: override `Map.merge_slashes` for this rule.

        :keyword bool build_only: set this to True and the rule will never match but will
                                  create a url that can be build. this is useful if you have
                                  resources on a subdomain or folder that are not handled by
                                  the WSGI application (like static data)

        :keyword str | callable redirect_to: if given this must be either a string
                                             or callable. in case of a callable it's
                                             called with the url adapter that
                                             triggered the match and the values
                                             of the url as keyword arguments and has
                                             to return the target for the redirect,
                                             otherwise it has to be a string with
                                             placeholders in rule syntax.

        :keyword bool alias: if enabled this rule serves as an alias for another rule with
                             the same endpoint and arguments.

        :keyword str host: if provided and the url map has host matching enabled this can be
                           used to provide a match rule for the whole host. this also means
                           that the subdomain feature is disabled.

        :keyword bool websocket: if set to True, this rule is only matches for
                                 websocket (`ws://`, `wss://`) requests. by default,
                                 rules will only match for http requests.
                                 defaults to False if not provided.

        :keyword int max_content_length: max content length that this route could handle,
                                         in bytes. if not provided, it will be set to
                                         `restricted_max_content_length` api config key.
                                         note that this value should be lesser than or equal
                                         to `max_content_length` api config key, otherwise
                                         it will cause an error.

        :keyword int status_code: status code to be returned on successful responses.
                                  defaults to corresponding status code of request's
                                  http method if not provided.

        :note status_code: it could be a value from `InformationResponseCodeEnum`
                           or `SuccessfulResponseCodeEnum` or `RedirectionResponseCodeEnum`.

        :keyword bool strict_status: specifies that it should only consider
                                     the status code as processed if it is from
                                     `InformationResponseCodeEnum` or
                                     `SuccessfulResponseCodeEnum` or
                                     `RedirectionResponseCodeEnum` values. otherwise
                                     all codes from `INFORMATION_CODE_MIN`
                                     to `INFORMATION_CODE_MAX` or from
                                     `SUCCESS_CODE_MIN` to `SUCCESS_CODE_MAX`
                                     or from `REDIRECTION_CODE_MIN` to
                                     `REDIRECTION_CODE_MAX` will be considered
                                     as processed. defaults to True if not provided.

        :keyword str | list[str] environments: a list of all environments that this
                                               route must be exposed on them.
                                               the values could be from all available
                                               environments in environments config store.
                                               for example: `production`, `development`.
                                               if not provided, the route will be exposed
                                               on all environments.

        :keyword ResultSchema | type[ResultSchema] result_schema: result schema to be used
                                                                  to filter results. it could
                                                                  be an instance or a type
                                                                  of `ResultSchema` class.

        :keyword bool indexed: specifies that list results must
                               include an extra field as row index.
                               the name of the index field and the initial value
                               of index could be provided by `index_name` and
                               `start_index` respectively. `indexed` keyword has
                               only effect if the returning result contains a list
                               of objects.

        :keyword str index_name: name of the extra field to contain
                                 the row index of each result. if not provided
                                 defaults to `row_num` value.

        :keyword int start_index: the initial value of row index. if not
                                  provided, starts from 1.

        :keyword SECURE_TRUE | SECURE_FALSE readable: specifies that any column or attribute
                                                      which has `allow_read=False` or its name
                                                      starts with underscore `_`, should not
                                                      be included in result dict. defaults
                                                      to `SECURE_TRUE` if not provided.
                                                      it will be used only for entity
                                                      conversion. this value will override
                                                      the corresponding value of
                                                      `result_schema` if provided.

        :keyword int depth: a value indicating the depth for conversion.
                            for example if entity A has a relationship with
                            entity B and there is a list of B in A, if `depth=0`
                            is provided, then just columns of A will be available
                            in result dict, but if `depth=1` is provided, then all
                            B entities in A will also be included in the result dict.
                            actually, `depth` specifies that relationships in an
                            entity should be followed by how much depth.
                            note that, if `columns` is also provided, it is required to
                            specify relationship property names in provided columns.
                            otherwise they won't be included even if `depth` is provided.
                            defaults to `default_depth` value of database config store.
                            please be careful on increasing `depth`, it could fail
                            application if set to higher values. choose it wisely.
                            normally the maximum acceptable `depth` would be 2 or 3.
                            there is a hard limit for max valid `depth` which is set
                            in `ConverterMixin.MAX_DEPTH` class variable. providing higher
                            `depth` value than this limit, will cause an error.
                            it will be used only for entity conversion.
                            this value will override the corresponding value of
                            `result_schema` if provided.

        :keyword bool no_cache: a value indicating that the response returning from this route
                                must have a `Cache-Control: no-cache` header. this header will
                                be automatically added. defaults to False if not provided.

        :keyword int request_limit: number of allowed requests to this
                                    route before it unregisters itself.
                                    defaults to None if not Provided.

        :keyword int lifetime: number of seconds in which this route must remain
                               responsive after initial registration. after this
                               period, the route will unregister itself.
                               defaults to None if not provided.

        :note request_limit, lifetime: if both of these values are provided, this
                                       route will unregister itself if any of
                                       these two conditions are met.

        :keyword bool paged: specifies that this route should return paginated results.
                             defaults to False if not provided.

        :keyword int page_size: default page size for this route.
                                defaults to `default_page_size` from
                                `database` config store if not provided.

        :keyword int max_page_size: maximum page size that client is allowed
                                    to request for this route. defaults to
                                    `max_page_size` from `database` configs store
                                    if not provided.

        :keyword bool cors_enabled: specifies that cross origin resource sharing is enabled.
                                    if not provided, it will be get from cors config store.

        :keyword bool cors_always_send: specifies that cors headers must be included in
                                        response even if the request does not have origin header.
                                        if not provided, it will be get from cors config store.

        :keyword list[str] cors_allowed_origins: a list of extra allowed origins to be used
                                                 in conjunction with default allowed ones.

        :keyword list[str] cors_exposed_headers: extra exposed headers to be combined
                                                 with default ones.

        :keyword list[str] cors_allowed_headers: extra allowed headers to be combined
                                                 with default ones.

        :keyword bool cors_allow_credentials: specifies that browsers are allowed to pass
                                              response headers to front-end javascript code
                                              if the route is authenticated.
                                              if not provided, it will be get from cors config
                                              store.

        :keyword int cors_max_age: maximum number of seconds to cache results.
                                   if not provided, it will be get from cors config store.

        :raises DuplicateRouteURLError: duplicate route url error.
        :raises OverwritingEndpointIsNotAllowedError: overwriting endpoint is not allowed error.
        :raises PageSizeLimitError: page size limit error.
        :raises MaxContentLengthLimitMismatchError: max content length limit mismatch error.
        :raises InvalidViewFunctionTypeError: invalid view function type error.
        :raises InvalidResultSchemaTypeError: invalid result schema type error.
        :raises InvalidResponseStatusCodeError: invalid response status code error.
        """

        application_services.add_url_rule(url, view_func=view_func,
                                          provide_automatic_options=provide_automatic_options,
                                          **options)



# -*- coding: utf-8 -*-
"""
validator decorators module.
"""

import pyrin.validator.services as validator_services


def validator(domain, field, *args, **kwargs):
    """
    decorator to register a validator.

    :param type[BaseEntity] | str domain: the domain in which this validator
                                          must be registered. it could be a
                                          type of a BaseEntity subclass.
                                          if a validator must be registered
                                          independent from any BaseEntity subclass,
                                          the domain could be a unique string name.
                                          note that the provided string name must be
                                          unique at application level.

    :param InstrumentedAttribute | str field: validator field name. it could be a
                                              string or a column. each validator will
                                              be registered with its field name in
                                              corresponding domain. to enable automatic
                                              validations, the provided field name must
                                              be the exact name of the parameter which
                                              this validator will validate. if you pass
                                              a column attribute, some constraints
                                              such as `nullable`, `min_length`, `max_length`,
                                              `min_value`, `max_value`, `allow_blank` and
                                              `allow_whitespace` could be extracted
                                              automatically from that column if not provided
                                              in inputs.

    :param object args: validator class constructor arguments.
    :param object kwargs: validator class constructor keyword arguments.

    :keyword bool replace: specifies that if there is another registered
                           validator with the same name and domain, replace
                           it with the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidValidatorTypeError: invalid validator type error.
    :raises DuplicatedValidatorError: duplicated validator error.

    :returns: validator class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available validator.

        :param type cls: validator class.

        :returns: validator class.
        :rtype: type
        """

        instance = cls(domain, field, *args, **kwargs)
        validator_services.register_validator(instance, **kwargs)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
security utils package.
"""

from pyrin.packaging.base import Package


class SecurityUtilsPackage(Package):
    """
    security utils package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'security.utils.component'



# -*- coding: utf-8 -*-
"""
transaction decorators module.
"""

from functools import update_wrapper

import pyrin.database.services as database_services


def atomic(*old_func, **options):
    """
    decorator to make a function execution atomic.

    meaning that before starting the execution of the function, a new session with a
    new transaction will be started, and after the completion of that function, if it
    was successful, the new transaction will be committed or if it was not successful
    the new transaction will be rolled back without the consideration or affecting the
    parent transaction which by default is scoped to request. the corresponding new
    session will also be removed after function execution.

    note that you *should not* commit, flush or rollback anything inside an atomic
    function, the `@atomic` decorator will handle commit, flush or rollback operations
    when needed. otherwise, unexpected behaviors may occur.

    also note that you *should not* remove the corresponding session from session factory
    when using `@atomic` decorator. the removal operation will be handled by decorator
    itself and if you remove session manually, it will cause broken chain of sessions
    and unexpected behaviour.

    this decorator also supports multiple `@atomic` usage in a single call hierarchy.
    for example:

    def service_root():
        store = get_current_store()
        value = EntityRoot()
        store.add(value)
        service_a()

    @atomic
    def service_a():
        store = get_current_store()
        value = EntityA()
        store.add(value)
        service_b()
        raise Exception()

    @atomic
    def service_b():
        store = get_current_store()
        value = EntityB()
        store.add(value)
        service_c()

    @atomic
    def service_c():
        value = EntityC()
        value.save()

    in the above example, if the call hierarchy starts with `service_root()`, at
    the end, the data of `service_c` and `service_b` will be persisted into database.
    but the data of `service_a` and `service_root` will not be persisted because
    `service_a` raises an error before finish.

    :param function old_func: function.

    :keyword bool expire_on_commit: expire atomic session after commit.
                                    it is useful to set it to True if
                                    the atomic function does not return
                                    any entities for post-processing.
                                    defaults to False if not provided.

    :returns: function result.
    """

    def decorator(func):
        """
        decorates the given function and makes its execution atomic.

        :param function func: function.

        :returns: decorated function
        """

        def wrapper(*args, **kwargs):
            """
            decorates the given function and makes its execution atomic.

            :param object args: function arguments.
            :param object kwargs: function keyword arguments.

            :returns: function result.
            """

            store = database_services.get_atomic_store(**options)
            try:
                result = func(*args, **kwargs)
                store.commit()
                return result
            except Exception as ex:
                store.rollback()
                raise ex
            finally:
                factory = database_services.get_current_session_factory()
                factory.remove(atomic=True)

        return update_wrapper(wrapper, func)

    if len(old_func) > 0:
        return decorator(old_func[0])

    return decorator


def nested(func):
    """
    decorator to execute a function in a nested transaction.

    meaning that before starting the execution of the function, a new nested transaction
    will be started, and after the completion of that function, if it was not successful
    the nested transaction will be rolled back without the consideration or affecting
    the parent transaction which by default is scoped to request.

    note that in nested transactions, the parent transaction must be committed to persist
    generated data during nested transaction, so committing the nested transaction itself,
    does not persist anything, it just releases the savepoint. if you want an independent
    transaction from parent that could commit its own changes on its own, use `@atomic`
    decorator.

    note that you *should not* commit, flush or rollback anything inside a nested
    function, the `@nested` decorator will handle commit, flush or rollback operations
    when needed. otherwise, unexpected behaviors may occur.

    :param function func: function.

    :returns: function result.
    """

    def decorator(*args, **kwargs):
        """
        decorates the given function and executes it in a nested transaction.

        :param object args: function arguments.
        :param object kwargs: function keyword arguments.

        :returns: function result.
        """

        store = database_services.get_current_store()
        nested_transaction = store.begin_nested()
        try:
            result = func(*args, **kwargs)
            nested_transaction.commit()
            return result
        except Exception as ex:
            nested_transaction.rollback()
            raise ex

    return update_wrapper(decorator, func)


def subtransaction(func):
    """
    decorator to execute a function in a subtransaction.

    meaning that before starting the execution of the function, a new subtransaction
    will be started, and after the completion of that function, if it was not successful
    the subtransaction will be rolled back and it will also force the parent transaction
    which by default is scoped to request to be rolled backed.

    note that in subtransactions, the parent transaction must be committed to persist
    generated data during subtransaction, so committing the subtransaction itself, does
    not persist anything, it just ends the subtransaction's scope. if you want an independent
    transaction from parent that could commit its own changes on its own, use `@atomic`
    decorator.

    note that you *should not* commit, flush or rollback anything inside a subtransaction
    function, the `@subtransaction` decorator will handle commit, flush or rollback
    operations when needed. otherwise, unexpected behaviors may occur.

    :param function func: function.

    :returns: function result.
    """

    def decorator(*args, **kwargs):
        """
        decorates the given function and executes it in a subtransaction.

        :param object args: function arguments.
        :param object kwargs: function keyword arguments.

        :returns: function result.
        """

        store = database_services.get_current_store()
        sub_transaction = store.begin(subtransactions=True)
        try:
            result = func(*args, **kwargs)
            sub_transaction.commit()
            return result
        except Exception as ex:
            sub_transaction.rollback()
            raise ex

    return update_wrapper(decorator, func)


def transient(func):
    """
    decorator to make a function execution transient.

    meaning that before starting the execution of the function, a new session with a
    new transaction will be started, and after the completion of that function, the
    new transaction will be rolled back without the consideration or affecting the
    parent transaction which by default is scoped to request. the corresponding new
    session will also be removed after function execution.

    note that you *should not* commit, flush or rollback anything inside a transient
    function, the `@transient` decorator will handle rollback operation when needed.
    otherwise, unexpected behaviors may occur.

    also note that you *should not* remove the corresponding session from session factory
    when using `@transient` decorator. the removal operation will be handled by decorator
    itself and if you remove session manually, it will cause broken chain of sessions
    and unexpected behaviour.

    this decorator also supports multiple `@transient` usage in a single call hierarchy.
    for example:

    def service_root():
        store = get_current_store()
        value = EntityRoot()
        store.add(value)
        service_a()

    @atomic
    def service_a():
        store = get_current_store()
        value = EntityA()
        store.add(value)
        service_b()

    @transient
    def service_b():
        store = get_current_store()
        value = EntityB()
        store.add(value)
        service_c()

    @transient
    def service_c():
        value = EntityC()
        value.save()

    in the above example, if the call hierarchy starts with `service_root()`, at
    the end, the data of `service_root` and `service_a` will be persisted into database.
    but the data of `service_b` and `service_c` will not be persisted because they are
    decorated as transient.

    :param function func: function.

    :returns: function result.
    """

    def decorator(*args, **kwargs):
        """
        decorates the given function and makes its execution transient.

        :param object args: function arguments.
        :param object kwargs: function keyword arguments.

        :returns: function result.
        """

        store = database_services.get_atomic_store()
        try:
            result = func(*args, **kwargs)
            return result
        finally:
            store.rollback()
            factory = database_services.get_current_session_factory()
            factory.remove(atomic=True)

    return update_wrapper(decorator, func)



# -*- coding: utf-8 -*-
"""
alembic exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class AlembicManagerException(CoreException):
    """
    alembic manager exception.
    """
    pass


class AlembicManagerBusinessException(CoreBusinessException,
                                      AlembicManagerException):
    """
    alembic manager business exception.
    """
    pass
